import json
from typing import Any
from typing import Literal
from typing import Union

from pydantic import Field
from pydantic import field_validator
from pydantic import model_validator

from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.cluster import Cluster
from laktory.models.resources.databricks.cluster import ClusterLibrary
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class JobCluster(Cluster):
    """
    Job Cluster. Same attributes as `laktory.models.Cluster`, except for

    * `access_controls`
    * `is_pinned`
    * `libraries`
    * `no_wait`

    that are not allowed.
    """

    access_controls: list[Any] = Field(None, exclude=True)
    is_pinned: bool = Field(None, exclude=True)
    libraries: list[Any] = Field(None, exclude=True)
    no_wait: bool = Field(None, exclude=True)

    @model_validator(mode="after")
    def excluded_fields(self) -> Any:
        for f in [
            "access_controls",
            "is_pinned",
            "libraries",
            "no_wait",
        ]:
            if getattr(self, f, None) not in [None, [], {}]:
                raise ValueError(f"Field {f} should be null")

        return self


class JobContinuous(BaseModel):
    """
    Job Continuous specifications

    Attributes
    ----------
    pause_status:
        Indicate whether this continuous job is paused or not. When the pause_status field is omitted in the block,
        the server will default to using `UNPAUSED` as a value for pause_status.
    """

    pause_status: Union[Literal["PAUSED", "UNPAUSED"], str] = None


class JobEmailNotifications(BaseModel):
    """
    Job Email Notifications specifications

    Attributes
    ----------
    no_alert_for_skipped_runs:
        If `True`, don't send alert for skipped runs. (It's recommended to use the corresponding setting in the
        notification_settings configuration block).
    on_duration_warning_threshold_exceededs:
        List of emails to notify when the duration of a run exceeds the threshold specified by the RUN_DURATION_SECONDS
        metric in the health block.
    on_failures:
        List of emails to notify when the run fails.
    on_starts:
        List of emails to notify when the run starts.
    on_successes:
        List of emails to notify when the run completes successfully.
    """

    no_alert_for_skipped_runs: bool = None
    on_duration_warning_threshold_exceededs: list[str] = None
    on_failures: list[str] = None
    on_starts: list[str] = None
    on_successes: list[str] = None

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "on_failures": "on_failure",
            "on_starts": "on_start",
            "on_successes": "on_success",
            "on_duration_warning_threshold_exceededs": "on_duration_warning_threshold_exceeded",
        }


class JobHealthRule(BaseModel):
    """
    Job Health Rule specifications

    Attributes
    ----------
    metric:
        Metric to check. The only supported metric is
        RUN_DURATION_SECONDS (check Jobs REST API documentation for the latest
        information).
    op:
        Operation used to compare operands. Currently, following operators are
        supported: EQUAL_TO, GREATER_THAN, GREATER_THAN_OR_EQUAL, LESS_THAN,
        LESS_THAN_OR_EQUAL, NOT_EQUAL.
    value:
        Value used to compare to the given metric.
    """

    metric: str = None
    op: str = None
    value: int = None


class JobHealth(BaseModel):
    """
    Job Health specifications

    Attributes
    ----------
    rules:
        Job health rules specifications
    """

    rules: list[JobHealthRule] = None


class JobNotificationSettings(BaseModel):
    """
    Job Notification Settings specifications

    Attributes
    ----------
    no_alert_for_canceled_runs:
        If `True`, don't send alert for cancelled runs.
    no_alert_for_skipped_runs:
        If `True`, don't send alert for skipped runs.
    """

    no_alert_for_canceled_runs: bool = None
    no_alert_for_skipped_runs: bool = None


class JobParameter(BaseModel):
    """
    Job Parameter specifications

    Attributes
    ----------
    default:
        Default value of the parameter.
    name:
        The name of the defined parameter. May only contain alphanumeric characters, `_`, `-`, and `.`,
    """

    default: str = None
    name: str = None


class JobRunAs(BaseModel):
    """
    Job Parameter specifications

    Attributes
    ----------
    service_principal_name:
        The application ID of an active service principal. Setting this field requires the servicePrincipal/user role.
    user_name:
        The email of an active workspace user. Non-admin users can only set this field to their own email.
    """

    service_principal_name: str = None
    user_name: str = None


class JobSchedule(BaseModel):
    """
    Job Schedule specifications

    Attributes
    ----------
    quartz_cron_expression:
        A Cron expression using Quartz syntax that describes the schedule for a job. This field is required.
    timezone_id:
        A Java timezone ID. The schedule for a job will be resolved with respect to this timezone. See Java TimeZone for
        details. This field is required.
    pause_status:
        Indicate whether this schedule is paused or not. When the pause_status field is omitted and a schedule is
        provided, the server will default to using `UNPAUSED` as a value for pause_status.
    """

    quartz_cron_expression: str
    timezone_id: str
    pause_status: Union[Literal["PAUSED", "UNPAUSED"], str, None] = None


class JobTaskConditionTask(BaseModel):
    """
    Job Task Condition Task specifications

    Attributes
    ----------
    left:
        The left operand of the condition task. It could be a string value, job state, or a parameter reference.
    op:
        The string specifying the operation used to compare operands. This task does not require a cluster to execute
         and does not support retries or notifications.
    right:
        The right operand of the condition task. It could be a string value, job state, or parameter reference.
    """

    left: str = None
    op: Literal[
        "EQUAL_TO",
        "GREATER_THAN",
        "GREATER_THAN_OR_EQUAL",
        "LESS_THAN",
        "LESS_THAN_OR_EQUAL",
        "NOT_EQUAL",
    ] = None
    right: str = None


class JobTaskDependsOn(BaseModel):
    """
    Job Task Depends On specifications

    Attributes
    ----------
    task_key:
        The name of the task this task depends on.
    outcome:
        Can only be specified on condition task dependencies. The outcome of the dependent task that must be met for
        this task to run.
    """

    task_key: str = None
    outcome: Literal["true", "false"] = None


class JobTaskNotebookTask(BaseModel):
    """
    Job Task Notebook Task specifications

    Attributes
    ----------
    notebook_path:
        The path of the databricks.Notebook to be run in the Databricks workspace or remote repository. For notebooks
        stored in the Databricks workspace, the path must be absolute and begin with a slash. For notebooks stored in a
        remote repository, the path must be relative.
    base_parameters:
        Base parameters to be used for each run of this job. If the run is initiated by a call to run-now with
        parameters specified, the two parameters maps will be merged. If the same key is specified in base_parameters
        and in run-now, the value from run-now will be used. If the notebook takes a parameter that is not specified
        in the job’s base_parameters or the run-now override parameters, the default value from the notebook will be
        used. Retrieve these parameters in a notebook using dbutils.widgets.get.
    warehouse_id:
        The id of the SQL warehouse to execute this task. If a warehouse_id is specified, that SQL warehouse will be
        used to execute SQL commands inside the specified notebook.
    source:
        Location type of the notebook, can only be WORKSPACE or GIT. When set to WORKSPACE, the notebook will be
        retrieved from the local Databricks workspace. When set to GIT, the notebook will be retrieved from a Git
        repository defined in git_source. If the value is empty, the task will use GIT if git_source is defined and
        WORKSPACE otherwise.
    """

    notebook_path: str
    base_parameters: dict[str, Any] = None
    warehouse_id: str = None
    source: Literal["WORKSPACE", "GIT"] = None


class JobTaskPipelineTask(BaseModel):
    """
    Job Task Pipeline specifications

    Attributes
    ----------
    pipeline_id:
        The pipeline's unique ID.
    full_refresh:
        Specifies if there should be full refresh of the pipeline.
    """

    pipeline_id: str = None
    full_refresh: bool = None


class JobTaskRunJobTask(BaseModel):
    """
    Job Task Run Job Task specifications

    Attributes
    ----------
    job_id:
        ID of the job
    job_parameters:
        Job parameters for the task
    """

    job_id: Union[int, str] = None
    job_parameters: dict[str, Any] = None


class JobTaskSqlTaskQuery(BaseModel):
    """
    Job Task SQL Task specifications

    Attributes
    ----------
    query_id:
        Query ID
    """

    query_id: str = None


class JobTaskSqlTaskAlertSubscription(BaseModel):
    """
    Job Task SQL Task Alert Subscription specifications

    Attributes
    ----------
    destination_id:

    user_name:
        The email of an active workspace user. Non-admin users can only set this field to their own email.
    """

    destination_id: str = None
    user_name: str = None


class JobTaskSQLTaskAlert(BaseModel):
    """
    Job Task SQL Task Alert specifications

    Attributes
    ----------
    alert_id:
        Identifier of the Databricks SQL Alert.
    subscriptions:
        A list of subscription blocks consisting out of one of the required fields: `user_name` for user emails or
        `destination_id` - for Alert destination's identifier.
    pause_subscriptions:
        It `True` subscriptions are paused
    """

    alert_id: str = None
    subscriptions: list[JobTaskSqlTaskAlertSubscription] = None
    pause_subscriptions: bool = None


class JobTaskSqlTaskDashboard(BaseModel):
    """
    Job Task SQL Task Dashboard specifications

    Attributes
    ----------
    dashboard_id:
        identifier of the Databricks SQL Dashboard databricks_sql_dashboard.
    custom_subject:
        Custom subject specifications
    subscriptions:
        Subscriptions specifications
    """

    dashboard_id: str = None
    custom_subject: list[JobTaskSqlTaskAlertSubscription] = None
    subscriptions: list[JobTaskSqlTaskAlertSubscription] = None


class JobTaskSqlTaskFile(BaseModel):
    """
    Job Task SQL Task File specifications

    Attributes
    ----------
    path:
        If source is `GIT`: Relative path to the file in the repository specified
        in the git_source block with SQL commands to execute. If source is
        `WORKSPACE`: Absolute path to the file in the workspace with SQL
        commands to execute.
    source:
        The source of the project. Possible values are `WORKSPACE` and `GIT`.
    """

    path: str = None
    source: Literal["WORKSPACE", "GIT"] = None


class JobTaskSQLTask(BaseModel):
    """
    Job Task SQL Task specifications

    Attributes
    ----------
    alert:
        Alert specifications
    dashboard:
        Dashboard specifications
    file:
        File specifications
    parameters:
        Parameters specifications
    query:
        Query specifications
    warehouse_id:
        Warehouse id
    """

    alert: JobTaskSQLTaskAlert = None
    dashboard: JobTaskSqlTaskDashboard = None
    file: JobTaskSqlTaskFile = None
    parameters: dict[str, Any] = None
    query: JobTaskSqlTaskQuery = None
    warehouse_id: str


class JobTaskForEachTaskTask(BaseModel):
    """
    Job Task specifications

    Attributes
    ----------
    condition_task:
        Condition Task specifications
    depends_ons:
        Depends On specifications
    description:
        specifications
    email_notifications:
        Email Notifications specifications
    existing_cluster_id:
        Cluster id from one of the clusters available in the workspace
    health:
        Job Health specifications
    job_cluster_key:
        Identifier that can be referenced in task block, so that cluster is shared between tasks
    libraries:
        Cluster Library specifications
    max_retries:
        An optional maximum number of times to retry an unsuccessful run.
    min_retry_interval_millis:
        An optional minimal interval in milliseconds between the start of the failed run and the subsequent retry run.
         The default behavior is that unsuccessful runs are immediately retried.
    notebook_task:
        Notebook Task specifications
    notification_settings:
        Notification Settings specifications
    pipeline_task:
        Pipeline Task specifications
    retry_on_timeout:
        If `True`, retry a job when it times out. The default behavior is to not retry on timeout.
    run_if:
        An optional value indicating the condition that determines whether the task should be run once its dependencies
        have been completed. When omitted, defaults to `ALL_SUCCESS`.
    run_job_task:
        Run Job specifications
    sql_task:
        SQL Task specifications
    task_key:
        A unique key for a given task.
    timeout_seconds:
        An optional timeout applied to each run of this job. The default behavior is to have no timeout.
    """

    # compute_key: str = None
    condition_task: JobTaskConditionTask = None
    depends_ons: list[JobTaskDependsOn] = None
    description: str = None
    email_notifications: JobEmailNotifications = None
    existing_cluster_id: str = None
    health: JobHealth = None
    job_cluster_key: str = None
    libraries: list[ClusterLibrary] = None
    max_retries: int = None
    min_retry_interval_millis: int = None
    # new_cluster: Cluster = None
    notebook_task: JobTaskNotebookTask = None
    notification_settings: JobNotificationSettings = None
    pipeline_task: JobTaskPipelineTask = None
    # python_wheel_task:
    retry_on_timeout: bool = None
    run_if: str = None
    run_job_task: JobTaskRunJobTask = None
    # spark_jar_task:
    # spark_python_task:
    sql_task: JobTaskSQLTask = None
    task_key: str = None
    timeout_seconds: int = None

    @field_validator("depends_ons")
    @classmethod
    def sort_depends_ons(cls, v: list[JobTaskDependsOn]) -> list[JobTaskDependsOn]:
        return sorted(v, key=lambda task: task.task_key)


class JobTaskForEachTask(BaseModel):
    """
    For Each Task specifications

    Attributes
    ----------
    inputs:
        Array for task to iterate on. This can be a JSON string or a reference
        to an array parameter. Laktory also supports a list input, which wil
        be serialized.
    task:
        Task to run against the inputs list.
    concurrency:
        Controls the number of active iteration task runs. Default is 20,
        maximum allowed is 100.
    """

    inputs: Union[str, list]
    task: JobTaskForEachTaskTask
    concurrency: int = None

    @field_validator("inputs")
    def parse_inputs(cls, v: Union[str, list]) -> str:
        if isinstance(v, list):
            v = json.dumps(v)

        v = v.replace("'", '"')

        return v


class JobTask(JobTaskForEachTaskTask):
    """
    Job Task specifications

    Attributes
    ----------
    for_each_task:
        For each task configuration
    """

    for_each_task: JobTaskForEachTask = None


class JobTriggerFileArrival(BaseModel):
    """
    Job Trigger File Arrival

    Attributes
    ----------
    url:
        URL of the job on the given workspace
    min_time_between_triggers_seconds:
        If set, the trigger starts a run only after the specified amount of
        time passed since the last time the trigger fired. The minimum allowed value is 60 seconds.
    wait_after_last_change_seconds:
        If set, the trigger starts a run only after no file activity has occurred for the specified amount of time.
        This makes it possible to wait for a batch of incoming files to arrive before triggering a run.
        The minimum allowed value is 60 seconds.
    """

    url: str = None
    min_time_between_triggers_seconds: int = None
    wait_after_last_change_seconds: int = None


class JobTrigger(BaseModel):
    """
    Job Trigger

    Attributes
    ----------
    file_arrival:
        File Arrival specifications
    pause_status:
        Indicate whether this trigger is paused or not. When the pause_status field is omitted in the block, the server
        will default to using `UNPAUSED` as a value for pause_status.
    """

    file_arrival: JobTriggerFileArrival
    pause_status: Union[Literal["PAUSED", "UNPAUSED"], str] = None


class JobWebhookNotificationsOnDurationWarningThresholdExceeded(BaseModel):
    """
    JobWebhook Notifications On Duration Warning Threshold specifications

    Attributes
    ----------
    id:
        Unique identifier
    """

    id: str = None


class JobWebhookNotificationsOnFailure(BaseModel):
    """
    JobWebhook Notifications On Failure specifications

    Attributes
    ----------
    id:
        Unique identifier
    """

    id: str = None


class JobWebhookNotificationsOnStart(BaseModel):
    """
    JobWebhook Notifications On Start specifications

    Attributes
    ----------
    id:
        Unique identifier
    """

    id: str = None


class JobWebhookNotificationsOnSuccess(BaseModel):
    """
    JobWebhook Notifications On Success specifications

    Attributes
    ----------
    id:
        Unique identifier
    """

    id: str = None


class JobWebhookNotifications(BaseModel):
    """
    Job Webhook Notifications specifications

    Attributes
    ----------
    on_duration_warning_threshold_exceededs:
        Warnings threshold exceeded specifications
    on_failures:
        On failure specifications
    on_starts:
        On starts specifications
    on_successes:
        On successes specifications
    """

    on_duration_warning_threshold_exceededs: list[
        JobWebhookNotificationsOnDurationWarningThresholdExceeded
    ] = None
    on_failures: list[JobWebhookNotificationsOnFailure] = None
    on_starts: list[JobWebhookNotificationsOnStart] = None
    on_successes: list[JobWebhookNotificationsOnSuccess] = None


class JobQueue(BaseModel):
    """
    Attributes
    ----------
    enabled:
        If true, enable queueing for the job.
    """

    enabled: bool


class JobLookup(ResourceLookup):
    """
    Attributes
    ----------
    id:
        The id of the databricks job
    """

    id: str = Field(serialization_alias="id")


class Job(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Job

    Attributes
    ----------
    access_controls:
        Access Controls specifications
    clusters:
        A list of job databricks.Cluster specifications that can be shared and reused by tasks of this job.
        Libraries cannot be declared in a shared job cluster. You must declare dependent libraries in task settings.
    continuous:
        Continuous specifications
    control_run_state:
        If `True`, the Databricks provider will stop and start the job as needed to ensure that the active run for the
        job reflects the deployed configuration. For continuous jobs, the provider respects the pause_status by
        stopping the current active run. This flag cannot be set for non-continuous jobs.
    description:
        An optional description for the job. The maximum length is 1024 characters in UTF-8 encoding.
    email_notifications:
        An optional set of email addresses notified when runs of this job begins, completes or fails. The default
        behavior is to not send any emails. This field is a block and is documented below.
    format:
    health:
        Health specifications
    lookup_existing:
        Specifications for looking up existing resource. Other attributes will
        be ignored.
    max_concurrent_runs:
        An optional maximum allowed number of concurrent runs of the job. Defaults to 1.
    max_retries:
        An optional maximum number of times to retry an unsuccessful run. A run is considered to be unsuccessful if it
        completes with a FAILED or INTERNAL_ERROR lifecycle state. The value -1 means to retry indefinitely and the
        value 0 means to never retry. The default behavior is to never retry. A run can have the following lifecycle
        state: PENDING, RUNNING, TERMINATING, TERMINATED, SKIPPED or INTERNAL_ERROR.
    min_retry_interval_millis:
        An optional minimal interval in milliseconds between the start of the failed run and the subsequent retry run.
        The default behavior is that unsuccessful runs are immediately retried.
    name:
        Name of the job
    name_prefix:
        Prefix added to the job name
    name_suffix:
        Suffix added to the job name
    notification_settings:
        Notifications specifications
    parameters:
        Parameters specifications
    retry_on_timeout:
        An optional policy to specify whether to retry a job when it times out. The default behavior is to not retry on
        timeout.
    run_as:
        Run as specifications
    schedule:
        Schedule specifications
    tags:
        Tags as key, value pairs
    tasks:
        Tasks specifications
    timeout_seconds:
        An optional timeout applied to each run of this job. The default behavior is to have no timeout.
    trigger:
        Trigger specifications
    webhook_notifications:
        Webhook notifications specifications

    Examples
    --------
    ```py
    import io

    from laktory import models

    # Define job
    job_yaml = '''
    name: job-stock-prices
    clusters:
      - name: main
        spark_version: 14.0.x-scala2.12
        node_type_id: Standard_DS3_v2

    tasks:
      - task_key: ingest
        job_cluster_key: main
        notebook_task:
          notebook_path: /jobs/ingest_stock_prices.py
        libraries:
          - pypi:
              package: yfinance

      - task_key: pipeline
        depends_ons:
          - task_key: ingest
        pipeline_task:
          pipeline_id: 74900655-3641-49f1-8323-b8507f0e3e3b

    access_controls:
      - group_name: account users
        permission_level: CAN_VIEW
      - group_name: role-engineers
        permission_level: CAN_MANAGE_RUN
    '''
    job = models.resources.databricks.Job.model_validate_yaml(io.StringIO(job_yaml))

    # Define job with for each task
    job_yaml = '''
    name: job-hello
    tasks:
      - task_key: hello-loop
        for_each_task:
          inputs:
            - id: 1
              name: olivier
            - id: 2
              name: kubic
          task:
            task_key: hello-task
            notebook_task:
              notebook_path: /Workspace/Users/olivier.soucy@okube.ai/hello-world
              base_parameters:
                input: "{{input}}"
    '''
    job = models.resources.databricks.Job.model_validate_yaml(io.StringIO(job_yaml))
    ```

    References
    ----------

    * [Databricks Job](https://docs.databricks.com/en/workflows/jobs/create-run-jobs.html)
    * [Pulumi Databricks Job](https://www.pulumi.com/registry/packages/databricks/api-docs/job/#databricks-job)
    """

    access_controls: list[AccessControl] = []
    clusters: list[JobCluster] = []
    continuous: JobContinuous = None
    control_run_state: bool = None
    description: str = None
    email_notifications: JobEmailNotifications = None
    format: str = None
    health: JobHealth = None
    lookup_existing: JobLookup = Field(None, exclude=True)
    max_concurrent_runs: int = None
    max_retries: int = None
    min_retry_interval_millis: int = None
    name: str = None
    name_prefix: str = None
    name_suffix: str = None
    notification_settings: JobNotificationSettings = None
    parameters: list[JobParameter] = []
    queue: JobQueue = None
    retry_on_timeout: bool = None
    run_as: JobRunAs = None
    schedule: JobSchedule = None
    tags: dict[str, Any] = {}
    tasks: list[JobTask] = []
    timeout_seconds: int = None
    trigger: JobTrigger = None
    webhook_notifications: JobWebhookNotifications = None

    @field_validator("tasks")
    @classmethod
    def sort_tasks(cls, v: list[JobTask]) -> list[JobTask]:
        return sorted(v, key=lambda task: task.task_key)

    @model_validator(mode="after")
    def update_name(self) -> Any:
        with self.validate_assignment_disabled():
            if self.name_prefix:
                self.name = self.name_prefix + self.name
                self.name_prefix = ""
            if self.name_suffix:
                self.name = self.name + self.name_suffix
                self.name_suffix = ""

        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    job_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Job"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls", "name_prefix", "name_suffix"]

    @property
    def pulumi_renames(self) -> dict[str, str]:
        return {"clusters": "job_clusters"}

    @property
    def pulumi_properties(self):
        d = super().pulumi_properties

        _clusters = []
        if self._camel_serialization:
            k = "jobClusters"
            if k in d:
                for c in d[k]:
                    name = c.pop("name")
                    _clusters += [
                        {
                            "jobClusterKey": name,
                            "newCluster": c,
                        }
                    ]
                d[k] = _clusters

        else:
            k = "job_clusters"
            if k in d:
                for c in d[k]:
                    name = c.pop("name")
                    _clusters += [
                        {
                            "job_cluster_key": name,
                            "new_cluster": c,
                        }
                    ]
                d[k] = _clusters

        return d

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_job"

    @property
    def terraform_renames(self) -> dict[str, str]:
        return {
            "clusters": "job_clusters",
            "cluster": "job_cluster",
        }

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes

    @property
    def terraform_properties(self) -> dict:
        d = super().terraform_properties

        _clusters = []
        k = "job_cluster"
        if k in d:
            for c in d[k]:
                name = c.pop("name")
                _clusters += [
                    {
                        "job_cluster_key": name,
                        "new_cluster": c,
                    }
                ]
            d[k] = _clusters

        return d



import importlib
import inspect
from functools import wraps

import pyspark.sql.session
from pyspark.sql.session import SparkSession

filepath = inspect.getabsfile(pyspark.sql.session)

spec = importlib.util.spec_from_file_location("spark0", filepath)
spark0 = importlib.util.module_from_spec(spec)
spec.loader.exec_module(spark0)


class LaktorySparkSession:
    """
    Databricks DLT module overwrite the spark.sql method, preventing from using
    the latest parametrized queries. To circumvent this, we read the function
    definition from the disk and assign it to a laktory namespace so that it
    can be used in addition to the overwritten sql.
    """

    def __init__(self, spark: SparkSession):
        self._spark = spark

    @wraps(spark0.SparkSession.sql)
    def sql(self, *args, **kwargs):
        return spark0.SparkSession.sql(self._spark, *args, **kwargs)


SparkSession.laktory: LaktorySparkSession = property(
    lambda self: LaktorySparkSession(self)
)



from typing import Literal
from typing import Union

from laktory._logger import get_logger
from laktory.models.transformers.basechain import BaseChain
from laktory.models.transformers.sparkchainnode import SparkChainNode

logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class SparkChain(BaseChain):
    """
    The `SparkChain` class defines a series of Spark transformation to be
    applied to a dataframe. Each transformation is expressed as a node
    (`SparkChainNode` object) that, upon execution, returns a new dataframe.
    Each node is executed sequentially in the provided order. A node may also
    be another `SparkChain`.

    Attributes
    ----------
    dataframe_backend:
        Differentiator to select dataframe chain type
    nodes:
        The list of transformations to be executed.

    Examples
    --------
    ```py
    import pandas as pd

    from laktory import models

    df0 = spark.createDataFrame(pd.DataFrame({"x": [1, 2, 3]}))

    # Build Chain
    sc = models.SparkChain(
        nodes=[
            {
                "with_column": {
                    "name": "cos_x",
                    "type": "double",
                    "expr": "F.cos('x')",
                },
            },
            {
                "nodes": [
                    {
                        "func_name": "withColumnRenamed",
                        "func_args": [
                            "x",
                            "x_tmp",
                        ],
                    },
                    {
                        "with_column": {
                            "name": "x2",
                            "type": "double",
                            "expr": "F.sqrt('x_tmp')",
                        },
                    },
                ],
            },
            {
                "func_name": "drop",
                "func_args": [
                    "x_tmp",
                ],
            },
        ]
    )

    # Execute Chain
    df = sc.execute(df0)

    # Print result
    print(df.toPandas().to_string())
    '''
          cos_x        x2
    0  0.540302  1.000000
    1 -0.416147  1.414214
    2 -0.989992  1.732051
    '''
    ```
    """

    dataframe_backend: Literal["SPARK"] = "SPARK"
    nodes: list[Union[SparkChainNode, "SparkChain"]]
    _view_definition: str = None
    _columns: list[list[str]] = []


SparkChain.model_rebuild()



from typing import Any
from typing import Literal

from pydantic import BaseModel
from pydantic import model_validator

from laktory._logger import get_logger
from laktory._settings import settings

logger = get_logger(__name__)


class PipelineChild(BaseModel):
    """ """

    dataframe_backend: Literal["SPARK", "POLARS"] = None

    @model_validator(mode="after")
    def update_children_after_init(self) -> Any:
        if not hasattr(self, "_parent"):
            self._parent = None
        self.update_children()
        return self

    def model_copy(self, *args, **kwargs):
        model = super().model_copy(*args, **kwargs)
        model.update_children()
        return model

    @property
    def child_attribute_names(self):
        return []

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value
        self.update_children()

    @property
    def df_backend(self) -> str:
        # Direct value
        backend = self.dataframe_backend
        if backend is not None:
            return backend

        # Value from parent
        parent = self._parent
        if parent is not None:
            return parent.df_backend

        # Value from settings
        return settings.dataframe_backend

    def update_children(self):
        for c_name in self.child_attribute_names:
            o = getattr(self, c_name)
            if o is None:
                continue
            elif isinstance(o, list):
                for _o in o:
                    _o.parent = self
            elif isinstance(o, dict):
                for _o in o.values():
                    _o.parent = self
            else:
                o.parent = self

    @property
    def parent_pipeline(self):
        from laktory.models.pipeline.pipeline import Pipeline

        def _get_pl(o):
            parent = getattr(o, "_parent", None)
            if parent is None:
                return None

            if isinstance(parent, Pipeline):
                return parent

            return _get_pl(parent)

        return _get_pl(self)

    @property
    def parent_pipeline_node(self):
        from laktory.models.pipeline.pipelinenode import PipelineNode

        def _get_pl(o):
            parent = getattr(o, "_parent", None)
            if parent is None:
                return None

            if isinstance(parent, PipelineNode):
                return parent

            return _get_pl(parent)

        return _get_pl(self)



from typing import Any

import polars as pl

from laktory._logger import get_logger

# from laktory.models.transformers.basechainnode import ChainNodeColumn


logger = get_logger(__name__)


def groupby_and_agg(
    df,
    groupby_columns: list[str] = None,
    agg_expressions: list[Any] = None,
) -> pl.DataFrame:
    """
    Apply a groupby and create aggregation columns.

    Parameters
    ----------
    df:
        DataFrame
    groupby_columns:
        List of column names to group by
    agg_expressions:
        List of columns defining the aggregations

    Examples
    --------
    ```py
    import polars as pl

    import laktory  # noqa: F401

    df0 = pl.DataFrame(
        {
            "symbol": ["AAPL", "AAPL"],
            "price": [200.0, 205.0],
            "tstamp": ["2023-09-01", "2023-09-02"],
        }
    )

    df = df0.laktory.groupby_and_agg(
        groupby_columns=["symbol"],
        agg_expressions=[
            {
                "name": "mean_price",
                "expr": "pl.col('price').mean()",
            },
        ],
    )

    print(df.glimpse(return_as_string=True))
    '''
    Rows: 1
    Columns: 2
    $ symbol     <str> 'AAPL'
    $ mean_price <f64> 202.5
    '''
    ```
    """
    from laktory.models.transformers.basechainnode import ChainNodeColumn

    # Parse inputs
    if agg_expressions is None:
        raise ValueError("`agg_expressions` must be specified")
    if groupby_columns is None:
        groupby_columns = []

    logger.info(f"Executing groupby ({groupby_columns}) with {agg_expressions}")

    # Groupby arguments
    groupby = []

    for c in groupby_columns:
        groupby += [c]

    # Agg arguments
    aggs = []
    for expr in agg_expressions:
        if not isinstance(expr, ChainNodeColumn):
            expr = ChainNodeColumn(**expr)

        expr.type = None
        aggs += [expr.eval(dataframe_backend="POLARS").alias(expr.name)]

    return df.group_by(groupby).agg(*aggs)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Literal
from typing import Union

from pydantic import field_validator
from pydantic import model_validator

from laktory._logger import get_logger
from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.dataquality.check import DataQualityCheck
from laktory.models.pipeline.orchestrators.databricksdltorchestrator import (
    DatabricksDLTOrchestrator,
)
from laktory.models.pipeline.orchestrators.databricksjoborchestrator import (
    DatabricksJobOrchestrator,
)
from laktory.models.pipeline.pipelinechild import PipelineChild
from laktory.models.pipeline.pipelinenode import PipelineNode
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource

if TYPE_CHECKING:
    import networkx as nx
    from plotly.graph_objs import Figure

logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Helper Classes                                                              #
# --------------------------------------------------------------------------- #


class PipelineUDF(BaseModel):
    """
    Pipeline User Define Function

    Attributes
    ----------
    module_name:
        Name of the module from which the function needs to be imported.
    function_name:
        Name of the function.
    module_path:
        Workspace filepath of the module, if not in the same directory as the pipeline notebook
    """

    module_name: str
    function_name: str
    module_path: str = None


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class Pipeline(BaseModel, PulumiResource, TerraformResource, PipelineChild):
    """
    Pipeline model to manage a full-fledged data pipeline including reading
    from data sources, applying data transformations through Spark and
    outputting to data sinks.

    A pipeline is composed of collections of `nodes`, each one defining its
    own source, transformations and optional sink. A node may be the source of
    another node.

    A pipeline may be run manually by using python or the CLI, but it may also
    be deployed and scheduled using one of the supported orchestrators, such as
    a Databricks Delta Live Tables or job.

    Attributes
    ----------
    databricks_job:
        Defines the Databricks Job specifications when DATABRICKS_JOB is
        selected as the orchestrator. Requires to add the supporting
        [notebook](https://github.com/okube-ai/laktory/blob/main/laktory/resources/quickstart-stacks/workflows/notebooks/jobs/job_laktory_pl.py)
        to the stack.
    databricks_dlt:
        Defines the Databricks DLT specifications when DATABRICKS_DLT is
        selected as the orchestrator. Requires to add the supporting
        [notebook](https://github.com/okube-ai/laktory/blob/main/laktory/resources/quickstart-stacks/workflows/notebooks/dlt/dlt_laktory_pl.py)
        to the stack.
    dependencies:
        List of dependencies required to run the pipeline. If Laktory is not
        provided, it's current version is added to the list.
    name:
        Name of the pipeline
    nodes:
        List of pipeline nodes. Each node defines a data source, a series
        of transformations and optionally a sink.
    orchestrator:
        Orchestrator used for scheduling and executing the pipeline. The
        selected option defines which resources are to be deployed.
        Supported options are:

        - `DATABRICKS_DLT`: When orchestrated through Databricks DLT, each
          pipeline node creates a DLT table (or view, if no sink is defined).
          Behind the scenes, `PipelineNodeDataSource` leverages native `dlt`
          `read` and `read_stream` functions to defined the interdependencies
          between the tables as in a standard DLT pipeline.
        - `DATABRICKS_JOB`: When deployed through a Databricks Job, a task
          is created for each pipeline node and all the required dependencies
          are set automatically. If a given task (or pipeline node) uses a
          `PipelineNodeDataSource` as the source, the data will be read from
          the upstream node sink.
    udfs:
        List of user defined functions provided to the transformer.
    root_path:
        Location of the pipeline node root used to store logs, metrics and
        checkpoints.

    Examples
    --------
    This first example shows how to configure a simple pipeline with 2 nodes.
    Upon execution, raw data will be read from a CSV file and two DataFrames
    (bronze and silver) will be created and saved as parquet files. Notice how
    the first node is used as a data source for the second node.

    ```py
    import io

    from laktory import models

    pipeline_yaml = '''
        name: pl-stock-prices
        nodes:
        - name: brz_stock_prices
          layer: BRONZE
          source:
            format: CSV
            path: ./raw/brz_stock_prices.csv
          sinks:
          - format: PARQUET
            mode: OVERWRITE
            path: ./dataframes/brz_stock_prices

        - name: slv_stock_prices
          layer: SILVER
          source:
            node_name: brz_stock_prices
          sinks:
          - format: PARQUET
            mode: OVERWRITE
            path: ./dataframes/slv_stock_prices
          transformer:
            nodes:
            - with_column:
                name: created_at
                type: timestamp
                expr: data.created_at
            - with_column:
                name: symbol
                expr: data.symbol
            - with_column:
                name: close
                type: double
                expr: data.close
            - func_name: drop
              func_args:
              - value: data
              - value: producer
              - value: name
              - value: description
    '''

    pl = models.Pipeline.model_validate_yaml(io.StringIO(pipeline_yaml))

    # Execute pipeline
    # pl.execute()
    ```

    The next example defines a 3 nodes pipeline (1 bronze and 2 silvers)
    orchestrated with a Databricks Job. Notice how nodes are used as data
    sources not only for other nodes, but also for the `other` keyword argument
    of the smart join function (slv_stock_prices). Because we are using the
    DATABRICKS_JOB orchestrator, the job configuration must be declared.
    The tasks will be automatically created by the Pipeline model. Each
    task will execute a single node using the notebook referenced in
    `databricks_job.notebook_path` the content of this notebook should be
    similar to laktory.resources.notebooks.job_laktory.pl

    ```py
    import io

    from laktory import models

    pipeline_yaml = '''
        name: pl-stock-prices
        orchestrator: DATABRICKS_JOB
        databricks_job:
          name: job-pl-stock-prices
          notebook_path: /Workspace/.laktory/jobs/job_laktory_pl.py
          clusters:
            - name: node-cluster
              spark_version: 14.0.x-scala2.12
              node_type_id: Standard_DS3_v2
        dependencies:
            - laktory==0.3.0
            - yfinance
        nodes:
        - name: brz_stock_prices
          layer: BRONZE
          source:
            path: /Volumes/dev/sources/landing/events/yahoo-finance/stock_price/
          sinks:
          -   path: /Volumes/dev/sources/landing/tables/dev_stock_prices/
              mode: OVERWRITE

        - name: slv_stock_prices
          layer: SILVER
          source:
            node_name: brz_stock_prices
          sinks:
          -   path: /Volumes/dev/sources/landing/tables/slv_stock_prices/
              mode: OVERWRITE
          transformer:
            nodes:
            - with_column:
                name: created_at
                type: timestamp
                expr: data.created_at
            - with_column:
                name: symbol
                expr: data.symbol
            - with_column:
                name: close
                type: double
                expr: data.close
            - func_name: drop
              func_args:
              - value: data
              - value: producer
              - value: name
              - value: description
            - func_name: laktory.smart_join
              func_kwargs:
                'on':
                  - symbol
                other:
                  node_name: slv_stock_meta

        - name: slv_stock_meta
          layer: SILVER
          source:
            path: /Volumes/dev/sources/landing/events/yahoo-finance/stock_meta/
          sinks:
          - path: /Volumes/dev/sources/landing/tables/slv_stock_meta/
            mode: OVERWRITE

    '''
    pl = models.Pipeline.model_validate_yaml(io.StringIO(pipeline_yaml))
    ```

    Finally, we re-implement the previous pipeline, but with a few key
    differences:

    - Orchestrator is `DATABRICKS_DLT` instead of a `DATABRICKS_JOB`
    - Sinks are Unity Catalog tables instead of storage locations
    - Data is read as a stream in most nodes
    - `slv_stock_meta` is simply a DLT view since it does not have an associated
      sink.

    We also need to provide some basic configuration for the DLT pipeline.

    ```py
    import io

    from laktory import models

    pipeline_yaml = '''
        name: pl-stock-prices
        orchestrator: DATABRICKS_DLT
        databricks_dlt:
            catalog: dev
            target: sandbox
            access_controls:
            - group_name: users
              permission_level: CAN_VIEW

        nodes:
        - name: brz_stock_prices
          layer: BRONZE
          source:
            path: /Volumes/dev/sources/landing/events/yahoo-finance/stock_price/
            as_stream: true
          sinks:
          - table_name: brz_stock_prices

        - name: slv_stock_prices
          layer: SILVER
          source:
            node_name: brz_stock_prices
            as_stream: true
          sinks:
          - table_name: slv_stock_prices
          transformer:
            nodes:
            - with_column:
                name: created_at
                type: timestamp
                expr: data.created_at
            - with_column:
                name: symbol
                expr: data.symbol
            - with_column:
                name: close
                type: double
                expr: data.close
            - func_name: drop
              func_args:
              - value: data
              - value: producer
              - value: name
              - value: description
            - func_name: laktory.smart_join
              func_kwargs:
                'on':
                  - symbol
                other:
                  node_name: slv_stock_meta

        - name: slv_stock_meta
          layer: SILVER
          source:
            path: /Volumes/dev/sources/landing/events/yahoo-finance/stock_meta/

    '''
    pl = models.Pipeline.model_validate_yaml(io.StringIO(pipeline_yaml))
    ```

    References
    ----------
    * [Databricks Job](https://docs.databricks.com/en/workflows/jobs/create-run-jobs.html)
    * [Databricks DLT](https://www.databricks.com/product/delta-live-tables)

    """

    databricks_job: Union[DatabricksJobOrchestrator, None] = None
    databricks_dlt: Union[DatabricksDLTOrchestrator, None] = None
    dataframe_backend: Literal["SPARK", "POLARS"] = None
    dependencies: list[str] = []
    name: str
    nodes: list[PipelineNode]
    orchestrator: Literal["DATABRICKS_DLT", "DATABRICKS_JOB", None] = None
    udfs: list[PipelineUDF] = []
    root_path: str = None

    @field_validator("root_path", mode="before")
    @classmethod
    def root_path_to_string(cls, value: Any) -> Any:
        if isinstance(value, Path):
            value = str(value)
        return value

    @model_validator(mode="before")
    @classmethod
    def assign_name(cls, data: Any) -> Any:
        for k in ["databricks_dlt", "databricks_job"]:
            o = data.get(k, None)
            if o and isinstance(o, dict):
                # orchestrator as a dict
                o["name"] = o.get("name", None) or data.get("name", None)
            elif o:
                # orchestrator as a model
                o.name = o.name or o.get("name", None)

        return data

    @model_validator(mode="before")
    @classmethod
    def push_df_backend(cls, data: Any) -> Any:
        """Need to push dataframe_backend which is required to differentiate between spark and polars transformer"""
        df_backend = data.get("dataframe_backend", None)
        if df_backend:
            if "nodes" in data.keys():
                for n in data["nodes"]:
                    if isinstance(n, dict):
                        n["dataframe_backend"] = n.get("dataframe_backend", df_backend)
                    else:
                        n.dataframe_backend = n.dataframe_backend or df_backend
        return data

    @model_validator(mode="after")
    def validate_orchestrator(self):
        if self.orchestrator == "DATABRICKS_JOB":
            if self.databricks_job is None:
                raise ValueError(
                    "databricks_job must be defined if DATABRICKS_JOB orchestrator is selected."
                )
            self.databricks_job.update_from_parent()

        if self.orchestrator == "DATABRICKS_DLT":
            if self.databricks_dlt is None:
                raise ValueError(
                    "databricks_job must be defined if DATABRICKS_DLT orchestrator is selected."
                )
            self.databricks_dlt.update_from_parent()

        return self

    # ----------------------------------------------------------------------- #
    # Children                                                                #
    # ----------------------------------------------------------------------- #

    @property
    def child_attribute_names(self):
        return ["nodes", "databricks_dlt", "databricks_job"]

    # ----------------------------------------------------------------------- #
    # ID                                                                      #
    # ----------------------------------------------------------------------- #

    @property
    def resolved_name(self) -> str:
        return self.inject_vars_into_dump({"name": self.name})["name"]

    @property
    def safe_name(self):
        name = self.resolved_name

        # Replace special characters
        chars = [" ", ".", "@", "{", "}", "[", "]", "$", "|"]
        for c in chars:
            name = name.replace(c, "-")

        return name

    # ----------------------------------------------------------------------- #
    # Libraries                                                               #
    # ----------------------------------------------------------------------- #

    @property
    def _dependencies(self):
        from laktory import __version__

        laktory_found = False

        dependencies = [d for d in self.dependencies]
        for d in dependencies:
            if "laktory" in d:
                laktory_found = True

        if not laktory_found:
            dependencies += [f"laktory=={__version__}"]

        return dependencies

    # ----------------------------------------------------------------------- #
    # Orchestrator                                                            #
    # ----------------------------------------------------------------------- #

    @property
    def is_orchestrator_dlt(self) -> bool:
        """If `True`, pipeline orchestrator is DLT"""
        return self.orchestrator == "DATABRICKS_DLT"

    # ----------------------------------------------------------------------- #
    # Paths                                                                   #
    # ----------------------------------------------------------------------- #

    @property
    def _root_path(self) -> Path:
        if self.root_path:
            return Path(self.root_path)

        return Path(settings.laktory_root) / "pipelines" / self.safe_name

    # ----------------------------------------------------------------------- #
    # Expectations                                                            #
    # ----------------------------------------------------------------------- #

    @property
    def checks(self) -> list[DataQualityCheck]:
        checks = []
        for node in self.nodes:
            checks += node.checks
        return checks

    # ----------------------------------------------------------------------- #
    # Nodes                                                                   #
    # ----------------------------------------------------------------------- #

    @property
    def nodes_dict(self) -> dict[str, PipelineNode]:
        """
        Nodes dictionary whose keys are the node names.

        Returns
        -------
        :
            Nodes
        """
        return {n.name: n for n in self.nodes}

    @property
    def dag(self) -> nx.DiGraph:
        """
        Networkx Directed Acyclic Graph representation of the pipeline. Useful
        to identify interdependencies between nodes.

        Returns
        -------
        :
            Directed Acyclic Graph
        """
        import networkx as nx

        # TODO: Review if this needs to be computed dynamically or if we can
        #       compute it only after initialization. It depends if we believe
        #       that the nodes will be changed after initialization.

        dag = nx.DiGraph()

        # Build nodes
        for n in self.nodes:
            dag.add_node(n.name)
        # Build edges and assign nodes to pipeline node data sources
        node_names = []
        for n in self.nodes:
            if n.name in node_names:
                raise ValueError(
                    f"Pipeline node '{n.name}' is declared twice in pipeline '{self.name}'"
                )
            node_names += [n.name]

            # for s in n.get_sources(PipelineNodeDataSource):
            for _node_name in n.upstream_node_names:
                dag.add_edge(_node_name, n.name)
                if _node_name not in self.nodes_dict:
                    raise ValueError(
                        f"Pipeline node data source '{_node_name}' is not defined in pipeline '{self.name}'"
                    )

        if not nx.is_directed_acyclic_graph(dag):
            for n in dag.nodes:
                logger.info(f"Pipeline {self.name} node: {n}")
            for e in dag.edges:
                logger.info(f"Pipeline {self.name} edge: {e[0]} -> {e[1]}")
            raise ValueError(
                f"Pipeline '{self.name}' is not a DAG (directed acyclic graph)."
                " A circular dependency has been detected. Please review nodes dependencies."
            )

        return dag

    @property
    def sorted_node_names(self):
        import networkx as nx

        return list(nx.topological_sort(self.dag))

    @property
    def sorted_nodes(self) -> list[PipelineNode]:
        """
        Topologically sorted nodes.

        Returns
        -------
        :
            List of Topologically sorted nodes.
        """
        node_names = self.sorted_node_names
        nodes = []
        for node_name in node_names:
            for node in self.nodes:
                if node.name == node_name:
                    nodes += [node]
                    break

        return nodes

    # ----------------------------------------------------------------------- #
    # Data Sources                                                            #
    # ----------------------------------------------------------------------- #

    @property
    def data_sources(self):
        sources = []
        for n in self.nodes:
            sources += n.data_sources
        return sources

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def purge(self, spark=None) -> None:
        logger.info("Purging Pipeline")

        for inode, node in enumerate(self.sorted_nodes):
            node.purge(
                spark=spark,
            )

    def execute(
        self, spark=None, udfs=None, write_sinks=True, full_refresh: bool = False
    ) -> None:
        """
        Execute the pipeline (read sources and write sinks) by sequentially
        executing each node. The selected orchestrator might impact how
        data sources or sinks are processed.

        Parameters
        ----------
        spark:
            Spark Session
        udfs:
            List of user-defined functions used in transformation chains.
        write_sinks:
            If `False` writing of node sinks will be skipped
        full_refresh:
            If `True` all nodes will be completely re-processed by deleting
            existing data and checkpoints before processing.
        """
        logger.info("Executing Pipeline")

        for inode, node in enumerate(self.sorted_nodes):
            node.execute(
                spark=spark,
                udfs=udfs,
                write_sinks=write_sinks,
                full_refresh=full_refresh,
            )

    def dag_figure(self) -> Figure:
        """
        [UNDER DEVELOPMENT] Generate a figure representation of the pipeline
        DAG.

        Returns
        -------
        :
            Plotly figure representation of the pipeline.
        """
        import networkx as nx
        import plotly.graph_objs as go

        dag = self.dag
        pos = nx.spring_layout(dag)

        # ------------------------------------------------------------------- #
        # Edges                                                               #
        # ------------------------------------------------------------------- #

        edge_traces = []
        for e in dag.edges:
            edge_traces += [
                go.Scatter(
                    x=[pos[e[0]][0], pos[e[1]][0]],
                    y=[pos[e[0]][1], pos[e[1]][1]],
                    line={
                        "color": "#a006b1",
                    },
                    marker=dict(
                        symbol="arrow-bar-up",
                        angleref="previous",
                        size=30,
                    ),
                    mode="lines+markers",
                    hoverinfo="none",
                    showlegend=False,
                )
            ]

        # ------------------------------------------------------------------- #
        # Nodes                                                               #
        # ------------------------------------------------------------------- #

        nodes_trace = go.Scatter(
            x=[_p[0] for _p in pos.values()],
            y=[_p[1] for _p in pos.values()],
            text=list(dag.nodes),
            name="pl nodes",
            mode="markers+text",
            marker=dict(
                size=50,
                color="#06d49e",
                line=dict(
                    width=2,
                    color="#dff2ed",
                ),
            ),
            textfont=dict(
                color="#a006b1",
            ),
        )

        return go.Figure(data=[nodes_trace] + edge_traces)

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def self_as_core_resources(self):
        return False

    @property
    def resource_type_id(self) -> str:
        """
        pl
        """
        return "pl"

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        if orchestrator is `DLT`:

        - DLT Pipeline

        if orchestrator is `DATABRICKS_JOB`:

        - Databricks Job
        """

        resources = []

        if self.databricks_job:
            resources += [self.databricks_job]

        if self.databricks_dlt:
            resources += [self.databricks_dlt]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return ""

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return ""



from laktory.dispatcher.dispatcher import Dispatcher
from laktory.dispatcher.dispatcherrunner import DispatcherRunner
from laktory.dispatcher.dltpipelinerunner import DLTPipelineRunner
from laktory.dispatcher.jobrunner import JobRunner



import copy
import json
import typing
from contextlib import contextmanager
from copy import deepcopy
from typing import Any
from typing import TextIO
from typing import TypeVar
from typing import Union
from typing import get_args
from typing import get_origin

import yaml  # TODO: Move into functions?
from pydantic import BaseModel as _BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import model_serializer
from pydantic._internal._model_construction import ModelMetaclass as _ModelMetaclass

from laktory._parsers import _resolve_value
from laktory._parsers import _resolve_values
from laktory._parsers import _snake_to_camel
from laktory.typing import var
from laktory.yaml.recursiveloader import RecursiveLoader

Model = TypeVar("Model", bound="BaseModel")


class ModelMetaclass(_ModelMetaclass):
    def __new__(
        mcs,
        cls_name: str,
        bases: tuple[type[Any], ...],
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> type:
        # Add var as a possible type hint of each model field to support variables injection
        for field_name in namespace.get("__annotations__", {}):
            type_hint = namespace["__annotations__"][field_name]

            if field_name.startswith("_"):
                continue

            if field_name in [
                "variables",
                "dataframe_backend",  # TODO: Review why this is required
            ]:
                continue

            if type_hint is None:
                continue

            if type_hint is typing.Any:
                continue

            origin = get_origin(type_hint)
            args = get_args(type_hint)
            new_type_hint = type_hint

            if origin is list:
                new_type_hint = list[tuple([Union[args[0], var]])]

            elif origin is dict:
                new_type_hint = dict[Union[args[0], var], Union[args[1], var]]

            new_type_hint = Union[new_type_hint, var]

            namespace["__annotations__"][field_name] = new_type_hint

        return super().__new__(mcs, cls_name, bases, namespace, **kwargs)


class BaseModel(_BaseModel, metaclass=ModelMetaclass):
    """
    Parent class for all Laktory models offering generic functions and
    properties. This `BaseModel` class is derived from `pydantic.BaseModel`.

    Attributes
    ----------
    variables:
        Variable values to be resolved when using `inject_vars` method.
    """

    model_config = ConfigDict(
        extra="forbid",
        # `validate_assignment` is required when injecting complex variables to resolve
        # target model and more suitable when models are dynamically updated in code.
        validate_assignment=True,
    )
    variables: dict[str, Any] = Field(default={}, exclude=True)
    _camel_serialization: bool = False
    _singular_serialization: bool = False

    @model_serializer(mode="wrap")
    def custom_serializer(self, handler) -> dict[str, Any]:
        dump = handler(self)
        if dump is None:
            return dump

        camel_serialization = self._camel_serialization
        singular_serialization = self._singular_serialization

        fields = {k: v for k, v in self.model_fields.items()}
        if camel_serialization:
            keys = list(dump.keys())
            for k in keys:
                k_camel = _snake_to_camel(k)
                if k_camel != k:
                    dump[_snake_to_camel(k)] = dump.pop(k)
                    fields[_snake_to_camel(k)] = fields[k]
                    k = _snake_to_camel(k)

                # TODO: Review and optimize (probably brittle)
                #       This is used to rename properties names from snake to
                #       camel keys when using Pulumi backend. Probably also
                #       need to parse list and dicts.
                if isinstance(dump[k], str) and dump[k].startswith("${resources."):
                    values = dump[k].split(".")
                    values[-1] = _snake_to_camel(values[-1])
                    dump[k] = ".".join(values)

        if singular_serialization:
            import inflect

            engine = inflect.engine()
            keys = list(dump.keys())
            for k in keys:
                if k in self.singularizations:
                    # Explicit singularization
                    k_singular = self.singularizations[k] or k
                else:
                    # Automatic singularization
                    k_singular = k
                    ann = str(fields[k].annotation)
                    if "list[typing.Union[laktory.models" in ann:
                        k_singular = engine.singular_noun(k) or k

                if k_singular != k:
                    dump[k_singular] = dump.pop(k)

        return dump

    # ----------------------------------------------------------------------- #
    # Class Methods                                                           #
    # ----------------------------------------------------------------------- #

    @classmethod
    def model_validate_yaml(cls, fp: TextIO) -> Model:
        """
        Load model from yaml file object using laktory.yaml.RecursiveLoader. Supports
        reference to external yaml and sql files using `!use`, `!extend` and `!update` tags.
        Path to external files can be defined using model or environment variables.

        Referenced path should always be relative to the file they are referenced from.

        Custom Tags
        -----------
        !use {filepath}:
            Directly inject the content of the file at `filepath`
        - !extend {filepath}:
            Extend the current list with the elements found in the file at `filepath`.
            Similar to python list.extend method.
        <<: !update {filepath}:
            Merge the current dictionary with the content of the dictionary defined at
            `filepath`. Similar to python dict.update method.

        Parameters
        ----------
        fp:
            file object structured as a yaml file

        Returns
        -------
        :
            Model instance

        Examples
        --------
        ```yaml
        businesses:
          apple:
            symbol: aapl
            address: !use addresses.yaml
            <<: !update common.yaml
            emails:
              - jane.doe@apple.com
              - extend! emails.yaml
          amazon:
            symbol: amzn
            address: !use addresses.yaml
            <<: update! common.yaml
            emails:
              - john.doe@amazon.com
              - extend! emails.yaml
        ```
        """

        data = RecursiveLoader.load(fp)
        return cls.model_validate(data)

    def model_dump_yaml(self, *args, **kwargs):
        return yaml.dump(self.model_dump(*args, **kwargs))

    @classmethod
    def model_validate_json_file(cls, fp: TextIO) -> Model:
        """
        Load model from json file object

        Parameters
        ----------
        fp:
            file object structured as a json file

        Returns
        -------
        :
            Model instance
        """
        data = json.load(fp)
        return cls.model_validate(data)

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def singularizations(self) -> dict[str, str]:
        return {}

    # ----------------------------------------------------------------------- #
    # Serialization                                                           #
    # ----------------------------------------------------------------------- #

    def _configure_serializer(self, camel=False, singular=False):
        self._camel_serialization = camel
        self._singular_serialization = singular
        for k in self.model_fields:
            f = getattr(self, k)
            if isinstance(f, BaseModel):
                f._configure_serializer(camel, singular)
            elif isinstance(f, list):
                for i in f:
                    if isinstance(i, BaseModel):
                        i._configure_serializer(camel, singular)
            elif isinstance(f, dict):
                for i in f.values():
                    if isinstance(i, BaseModel):
                        i._configure_serializer(camel, singular)

    # ----------------------------------------------------------------------- #
    # Validation ByPass                                                       #
    # ----------------------------------------------------------------------- #

    @contextmanager
    def validate_assignment_disabled(self):
        """
        Updating a model attribute inside a model validator when `validate_assignment`
        is `True` causes an infinite recursion by design and must be turned off
        temporarily.
        """
        original_state = self.model_config["validate_assignment"]
        self.model_config["validate_assignment"] = False
        try:
            yield
        finally:
            self.model_config["validate_assignment"] = original_state

    # ----------------------------------------------------------------------- #
    # Variables Injection                                                     #
    # ----------------------------------------------------------------------- #

    def push_vars(self, update_core_resources=False) -> Any:
        """Push variable values to all child recursively"""

        def _update_model(m):
            if not isinstance(m, BaseModel):
                return
            for k, v in self.variables.items():
                m.variables[k] = m.variables.get(k, v)
            m.push_vars()

        def _push_vars(o):
            if isinstance(o, list):
                for _o in o:
                    _push_vars(_o)
            elif isinstance(o, dict):
                for _o in o.values():
                    _push_vars(_o)
            else:
                _update_model(o)

        for k in self.model_fields.keys():
            _push_vars(getattr(self, k))

        if update_core_resources and hasattr(self, "core_resources"):
            for r in self.core_resources:
                if r != self:
                    _push_vars(r)

        return None

    def inject_vars(self, inplace: bool = False, vars: dict = None):
        """
        Inject model variables values into a model attributes.

        Parameters
        ----------
        inplace:
            If `True` model is modified in place. Otherwise, a new model
            instance is returned.
        vars:
            A dictionary of variables to be injected in addition to the
            model internal variables.


        Returns
        -------
        :
            Model instance.

        Examples
        --------
        ```py
        from typing import Union

        from laktory import models


        class Cluster(models.BaseModel):
            name: str = None
            size: Union[int, str] = None


        c = Cluster(
            name="cluster-${vars.my_cluster}",
            size="${{ 4 if vars.env == 'prod' else 2 }}",
            variables={
                "env": "dev",
            },
        ).inject_vars()
        print(c)
        # > variables={'env': 'dev'} name='cluster-${vars.my_cluster}' size=2
        ```

        References
        ----------
        * [variables](https://www.laktory.ai/concepts/variables/)
        """

        # Fetching vars
        if vars is None:
            vars = {}
        vars = deepcopy(vars)
        vars.update(self.variables)

        # Create copy
        if not inplace:
            self = self.model_copy(deep=True)

        # Inject into field values
        for k in self.model_fields_set:
            if k == "variables":
                continue
            o = getattr(self, k)

            if isinstance(o, BaseModel) or isinstance(o, dict) or isinstance(o, list):
                # Mutable objects will be updated in place
                _resolve_values(o, vars)
            else:
                # Simple objects must be updated explicitly
                setattr(self, k, _resolve_value(o, vars))

        if not inplace:
            return self

    def inject_vars_into_dump(
        self, dump: dict[str, Any], inplace: bool = False, vars: dict[str, Any] = None
    ):
        """
        Inject model variables values into a model dump.

        Parameters
        ----------
        dump:
            Model dump (or any other general purpose mutable object)
        inplace:
            If `True` model is modified in place. Otherwise, a new model
            instance is returned.
        vars:
            A dictionary of variables to be injected in addition to the
            model internal variables.


        Returns
        -------
        :
            Model dump with injected variables.


        Examples
        --------
        ```py
        from laktory import models

        m = models.BaseModel(
            variables={
                "env": "dev",
            },
        )
        data = {
            "name": "cluster-${vars.my_cluster}",
            "size": "${{ 4 if vars.env == 'prod' else 2 }}",
        }
        print(m.inject_vars_into_dump(data))
        # > {'name': 'cluster-${vars.my_cluster}', 'size': 2}
        ```

        References
        ----------
        * [variables](https://www.laktory.ai/concepts/variables/)
        """

        # Setting vars
        if vars is None:
            vars = {}
        vars = deepcopy(vars)
        vars.update(self.variables)

        # Create copy
        if not inplace:
            dump = copy.deepcopy(dump)

        # Inject into field values
        _resolve_values(dump, vars)

        if not inplace:
            return dump



import polars as pl

__all__ = [
    # "coalesce",
    "row_number",
]


# def coalesce(exprs: Union[str, pl.Expr, list[str], list[pl.Expr]], *more_exprs, available_columns=None):
#
#     if not available_columns:
#         return pl.coalesce(exprs, *more_exprs)
#
#     # Parse inputs
#     if not isinstance(exprs, list):
#         exprs = [exprs]
#     exprs = [e for e in exprs] + [e for e in more_exprs]
#
#     _exprs = []
#     for e in exprs:
#         if isinstance(e, str) and e in available_columns:
#             _exprs += [e]
#
#         elif isinstance(e, pl.Expr):
#             found = all([c in available_columns for c in e.meta.root_names()])
#             if found:
#                 _exprs += [e]
#
#     return pl.coalesce(_exprs)


def row_number() -> pl.Expr:
    """
    Window function: returns a sequential number starting at 1 within a window partition.

    Returns
    -------
    :
        Output column

    Examples
    --------
    ```py
    import polars as pl

    import laktory  # noqa: F401

    df = pl.DataFrame(
        {
            "x": ["a", "a", "b", "b", "b", "c"],
            "z": ["11", "10", "22", "21", "20", "30"],
        }
    )
    df = df.with_columns(y1=pl.Expr.laktory.row_number())
    df = df.with_columns(y2=pl.Expr.laktory.row_number().over("x"))
    df = df.with_columns(y3=pl.Expr.laktory.row_number().sort_by("z").over("x"))
    print(df.glimpse(return_as_string=True))
    '''
    Rows: 6
    Columns: 5
    $ x  <str> 'a', 'a', 'b', 'b', 'b', 'c'
    $ z  <str> '11', '10', '22', '21', '20', '30'
    $ y1 <u32> 1, 2, 3, 4, 5, 6
    $ y2 <u32> 1, 2, 1, 2, 3, 1
    $ y3 <u32> 2, 1, 3, 2, 1, 1
    '''
    ```
    """
    return pl.int_range(1, pl.len() + 1, dtype=pl.UInt32)



from typing import Any
from typing import Union

from pydantic import model_validator

from laktory._logger import get_logger
from laktory.models.datasources.basedatasource import BaseDataSource
from laktory.polars import PolarsLazyFrame
from laktory.polars import is_polars_dataframe
from laktory.spark import SparkDataFrame
from laktory.spark import is_spark_dataframe

logger = get_logger(__name__)


class MemoryDataSource(BaseDataSource):
    """
    Data source using in-memory DataFrame, generally used in the context of a
    data pipeline.

    Attributes
    ----------
    data:
        Serialized data to build input DataFrame
    df:
        Input DataFrame

    Examples
    ---------
    ```python
    import polars as pl

    from laktory import models

    data = {
        "symbol": ["AAPL", "GOOGL"],
        "price": [200.0, 205.0],
        "tstamp": ["2023-09-01", "2023-09-01"],
    }

    # Spark from dict
    source = models.MemoryDataSource(
        data=data,
        dataframe_backend="SPARK",
    )
    df = source.read(spark=spark)
    print(df.laktory.show_string())
    '''
    +-----+------+----------+
    |price|symbol|    tstamp|
    +-----+------+----------+
    |200.0|  AAPL|2023-09-01|
    |205.0| GOOGL|2023-09-01|
    +-----+------+----------+
    '''

    # Polars from df
    source = models.MemoryDataSource(
        df=pl.DataFrame(data),
    )
    df = source.read()
    print(df.to_pandas())
    '''
      symbol  price      tstamp
    0   AAPL  200.0  2023-09-01
    1  GOOGL  205.0  2023-09-01
    '''
    ```
    """

    data: Union[dict[str, list[Any]], list[dict[str, Any]]] = None
    df: Any = None

    @model_validator(mode="after")
    def validate_input(self) -> Any:
        if self.df is None and self.data is None:
            raise ValueError("Either `data` or `df` must be provided.")

        if self.df is not None and self.data is not None:
            raise ValueError("Only `data` or `df` can be provided.")

        if self.df is not None:
            if is_spark_dataframe(self.df):
                dataframe_backend = "SPARK"
            elif is_polars_dataframe(self.df):
                dataframe_backend = "POLARS"
            else:
                raise ValueError("DataFrame must be of type Spark or Polars")

            with self.validate_assignment_disabled():
                self.dataframe_backend = dataframe_backend

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    #
    # @property
    # def _id(self):
    #     return str(self.path)

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _read_spark(self, spark) -> SparkDataFrame:
        logger.info(f"Reading {self._id} from memory")

        if self.df is not None:
            return self.df

        data = self.data
        if isinstance(data, dict):
            data = [dict(zip(data.keys(), values)) for values in zip(*data.values())]

        return spark.createDataFrame(data)

    def _read_polars(self) -> PolarsLazyFrame:
        logger.info(f"Reading {self._id} from memory")

        import polars as pl

        if self.df is not None:
            return self.df

        return pl.LazyFrame(self.data)



from functools import wraps

from pyspark.sql.dataframe import DataFrame

from laktory.spark.dataframe.display import display
from laktory.spark.dataframe.groupby_and_agg import groupby_and_agg
from laktory.spark.dataframe.has_column import has_column
from laktory.spark.dataframe.schema_flat import schema_flat
from laktory.spark.dataframe.show_string import show_string
from laktory.spark.dataframe.smart_join import smart_join
from laktory.spark.dataframe.watermark import watermark
from laktory.spark.dataframe.window_filter import window_filter


class LaktoryDataFrame:
    def __init__(self, df: DataFrame):
        self._df = df

    @wraps(display)
    def display(self, *args, **kwargs):
        return display(self._df, *args, **kwargs)

    @wraps(groupby_and_agg)
    def groupby_and_agg(self, *args, **kwargs):
        return groupby_and_agg(self._df, *args, **kwargs)

    @wraps(has_column)
    def has_column(self, *args, **kwargs):
        return has_column(self._df, *args, **kwargs)

    @wraps(schema_flat)
    def schema_flat(self, *args, **kwargs):
        return schema_flat(self._df, *args, **kwargs)

    @wraps(show_string)
    def show_string(self, *args, **kwargs):
        return show_string(self._df, *args, **kwargs)

    @wraps(smart_join)
    def smart_join(self, *args, **kwargs):
        return smart_join(self._df, *args, **kwargs)

    @wraps(watermark)
    def watermark(self, *args, **kwargs):
        return watermark(self._df, *args, **kwargs)

    @wraps(window_filter)
    def window_filter(self, *args, **kwargs):
        return window_filter(self._df, *args, **kwargs)


DataFrame.laktory: LaktoryDataFrame = property(lambda self: LaktoryDataFrame(self))


# Spark Connect DataFrame Extensions
try:
    from pyspark.sql.connect.dataframe import DataFrame

    DataFrame.laktory = property(lambda self: LaktoryDataFrame(self))
except:  # noqa: E722
    pass



from typing import Literal
from typing import Union

from laktory.models.basemodel import BaseModel


class DataProducer(BaseModel):
    """
    Specifications on the producer of data.

    Attributes
    ----------
    name:
        Name of the data producer
    description:
        Description of the data producer
    party:
        First, second or third party
        ref.: https://blog.hubspot.com/service/first-party-data

    Examples
    --------
    ```py
    from laktory import models

    producer = models.DataProducer(
        name="yahoo-finance",
        description="yfinance offers a threaded and Pythonic way to download market data from Yahoo! finance",
        party=3,
    )
    print(producer)
    '''
    variables={} name='yahoo-finance' description='yfinance offers a threaded and Pythonic way to download market data from Yahoo! finance' party=3
    '''
    ```
    """

    name: str
    description: Union[str, None] = None
    party: Literal[1, 2, 3] = 1



from typing import Literal
from typing import Union

from pydantic import Field
from pydantic import model_validator

from laktory.models.basemodel import BaseModel
from laktory.models.grants.cataloggrant import CatalogGrant
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.grants import Grants
from laktory.models.resources.databricks.schema import Schema
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class CatalogLookup(ResourceLookup):
    """
    Attributes
    ----------
    name:
        Catalog name
    """

    name: str = Field(serialization_alias="id")


class Catalog(BaseModel, PulumiResource, TerraformResource):
    """
    A catalog is the first layer of Unity Catalog’s three-level namespace. It’s
    used to organize your data assets.

    Attributes
    ----------
    comment:
        Text description of the catalog
    force_destroy:
        If `True` catalog can be deleted, even when not empty
    grants:
        List of grants operating on the catalog
    isolation_mode:
        Whether the catalog is accessible from all workspaces or a specific set
        of workspaces. Can be ISOLATED or OPEN. Setting the catalog to ISOLATED
        will automatically allow access from the current workspace.
    lookup_existing:
        Specifications for looking up existing resource. Other attributes will
        be ignored.
    name:
        Name of the catalog
    owner:
        User/group/service principal name of the catalog owner
    schemas:
        List of schemas stored in the catalog
    storage_root:
        Managed location of the catalog. Location in cloud storage where data
        for managed tables will be stored. If not specified, the location will
        default to the metastore root location.

    Examples
    --------
    ```py
    from laktory import models

    catalog = models.resources.databricks.Catalog(
        name="dev",
        grants=[
            {"principal": "account users", "privileges": ["USE_CATALOG", "USE_SCHEMA"]}
        ],
        schemas=[
            {
                "name": "engineering",
                "grants": [
                    {"principal": "domain-engineering", "privileges": ["SELECT"]}
                ],
            },
            {
                "name": "sources",
                "volumes": [
                    {
                        "name": "landing",
                        "volume_type": "EXTERNAL",
                        "grants": [
                            {
                                "principal": "account users",
                                "privileges": ["READ_VOLUME"],
                            },
                            {
                                "principal": "role-metastore-admins",
                                "privileges": ["WRITE_VOLUME"],
                            },
                        ],
                    },
                ],
            },
        ],
    )
    ```

    References
    ----------

    * [Databricks Unity Catalog](https://docs.databricks.com/en/data-governance/unity-catalog/index.html#catalogs)
    * [Pulumi Databricks Catalog](https://www.pulumi.com/registry/packages/databricks/api-docs/catalog/)
    """

    comment: Union[str, None] = None
    force_destroy: bool = True
    grants: list[CatalogGrant] = None
    isolation_mode: Literal["OPEN", "ISOLATED"] = "OPEN"
    lookup_existing: CatalogLookup = Field(None, exclude=True)
    name: str
    owner: str = None
    schemas: list[Schema] = []
    storage_root: str = None

    @model_validator(mode="after")
    def assign_name(self):
        for schema in self.schemas:
            schema.catalog_name = self.name
            schema.assign_name()

        return self

    # ----------------------------------------------------------------------- #
    # Computed fields                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def full_name(self) -> str:
        """Full name of the catalog `{catalog_name}`"""
        return self.name

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - catalog grants
        - schemas resources
        """
        resources = []

        # Catalog grants
        if self.grants:
            grant = Grants(
                resource_name=f"grants-{self.resource_name}",
                catalog=self.full_name,
                grants=[
                    {"principal": g.principal, "privileges": g.privileges}
                    for g in self.grants
                ],
            )
            resources += [grant]

        # Catalog schemas
        if self.schemas:
            for s in self.schemas:
                resources += s.core_resources

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Catalog"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["schemas", "is_unity", "grants"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_catalog"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from .basemodel import BaseModel
from .dataevent import DataEvent
from .dataframecolumnexpression import DataFrameColumnExpression
from .dataproducer import DataProducer
from .dataquality import *
from .datasinks import *
from .datasources import *
from .grants import *
from .pipeline.pipeline import Pipeline
from .pipeline.pipelinenode import PipelineNode
from .resources import *
from .stacks import *
from .transformers import *



from typing import Union

import pyspark.sql.functions as F
from pyspark.sql.column import Column

COLUMN_OR_NAME = Union[Column, str]
"""spark column or column name"""

INT_OR_COLUMN = Union[int, Column, str]
"""int, spark column or column name"""

FLOAT_OR_COLUMN = Union[float, Column, str]
"""float, spark column or column name"""

STRING_OR_COLUMN = Union[str, Column]
"""string or spark column"""


def _col(col: str) -> Column:
    if isinstance(col, Column):
        return col

    return F.expr(col)


def _lit(col: str) -> Column:
    if isinstance(col, Column):
        return col

    return F.lit(col)



from laktory._logger import get_logger
from laktory.spark import SparkDataFrame

try:
    from dlt import *
    from dlt import apply_changes as _apply_changes
    from dlt import read as _read
    from dlt import read_stream as _read_stream
except Exception:
    pass


logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Utilities                                                                   #
# --------------------------------------------------------------------------- #


def is_mocked() -> bool:
    """
    DLT module mock flag. If `True`, native Databricks DLT functions are mocked
    to allow debugging by running a pipeline notebook against an arbitrary
    cluster even when the native DLT module is not available.

    DLT is mocked in the following situations:

    * notebook ran outside a DLT pipeline and DBR < 13
    * notebook ran outside a DLT pipeline and DBR >= 13 with shared access
    mode cluster

    Returns
    -------
    :
        Mocked flag
    """
    try:
        import dlt

        return False
    except Exception:
        return True


def is_debug() -> bool:
    """
    Debug flag. If `True`, DLT readers are replaced with laktory functions
    allowing to run a notebook outside of a pipeline and preview the content
    of the output dataframe.

    Returns
    -------
    :
        Debug flag
    """
    try:
        import dlt
    except Exception:
        return True
    if spark.conf.get("pipelines.dbrVersion", None) is None:
        return True
    return False


def get_df(df_wrapper) -> SparkDataFrame:
    """
    When executed in debug mode (ref `dlt.is_debug`), executes and returns
    the output dataframe of a function decorated with `@dlt.table` or
    `@dlt.view`. Returns None when `is_debug()` is `False`.

    This method is not supported when a table using DLT views as input.

    Returns
    -------
    :
        Output dataframe

    Examples
    --------
    ```py
    from laktory import dlt

    dlt.spark = spark


    def define_table():
        @dlt.table(name="stock_prices")
        def get_df():
            df = spark.createDataFrame([[1.0, 1.1, 0.98]])
            return df

        return get_df


    wrapper = define_table()
    df = dlt.get_df(wrapper)
    display(df)
    ```
    """
    df = None
    if is_debug():
        if is_mocked():
            df = df_wrapper()
        else:
            df = df_wrapper().func()
    return df


# --------------------------------------------------------------------------- #
# Overwrite                                                                   #
# --------------------------------------------------------------------------- #


def read(*args, **kwargs) -> SparkDataFrame:
    """
    When `is_debug()` is `True` read table from storage, else read table from
    pipeline with native Databricks `dlt.read`

    Returns
    -------
    :
        Ouput dataframe

    Examples
    --------
    ```py
    from laktory import dlt

    dlt.spark = spark


    def define_table():
        @dlt.table(name="slv_stock_prices")
        def get_df():
            df = dlt.read("dev.finance.brz_stock_prices")
            return df

        return get_df


    define_table()
    ```
    """
    if is_debug():
        return spark.read.table(args[0])
    else:
        # Remove catalog and schema from naming space
        args = list(args)
        args[0] = args[0].split(".")[-1]
        return _read(*args, **kwargs)


def read_stream(*args, fmt="DELTA", **kwargs):
    """
    When `is_debug()` is `True` read table from storage as stream, else read
    table from pipeline with native Databricks `dlt.read_stream`

    Returns
    -------
    :
        Ouput dataframe

    Examples
    --------
    ```py
    from laktory import dlt

    dlt.spark = spark


    def define_table():
        @dlt.table(name="slv_stock_prices")
        def get_df():
            df = dlt.read_stream("dev.finance.brz_stock_prices")
            return df

        return get_df


    define_table()
    ```
    """

    if is_debug():
        return spark.readStream.format(fmt).table(args[0])
    else:
        # Remove catalog and schema from naming space
        args = list(args)
        args[0] = args[0].split(".")[-1]
        return _read_stream(*args, **kwargs)


def apply_changes(*args, node=None, **kwargs):
    """
    When `is_debug()` is `True` read source CDC table from storage, else run
    native Databricks `dlt.apply_changes`

    Returns
    -------
    :
        Output dataframe

    Examples
    --------
    ```py
    from laktory import dlt
    from laktory import models

    dlt.spark = spark


    def define_table(node, sink):
        dlt.create_streaming_table(name=sink.table_name)
        df = dlt.apply_changes(
            source=node.source.table_name, **sink.dlt_apply_changes_kwargs
        )
        return df


    node = models.PipelineNode(
        name="slv_stock_prices",
        source={
            "table_name": "brz_stock_prices",
        },
        sinks=[
            {
                "table_name": "brz_stock_prices",
                "mode": "MERGE",
                "merge_cdc_options": {
                    "primary_keys": ["asset_symbol"],
                    "order_by": "change_id",
                    "scd_type": 2,
                },
            }
        ],
    )

    define_table(node, node.primary_sink)
    ```
    """
    if is_debug():
        if node is None:
            return
        df = node.source.read(spark=spark)
        # TODO: Apply changes
        logger.warning(
            "Laktory does not currently support applying CDC changes. Returned dataframe is CDC source."
        )
        return df
    else:
        return _apply_changes(*args, **kwargs)


# --------------------------------------------------------------------------- #
# Mocks                                                                       #
# --------------------------------------------------------------------------- #

if is_mocked():

    def table(*table_args, **table_kwargs):
        def decorator(func):
            def wrapper(*args, **kwargs):
                logger.info(f"Running {func.__name__}")
                return func(*args, **kwargs)

            return wrapper

        return decorator

    def view(*view_args, **view_kwargs):
        def decorator(func):
            def wrapper(*args, **kwargs):
                logger.info(f"Running {func.__name__}")
                return func(*args, **kwargs)

            return wrapper

        return decorator

    def expect(*table_args, **table_kwargs):
        def decorator(func):
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return decorator

    def expect_or_drop(*table_args, **table_kwargs):
        def decorator(func):
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return decorator

    def expect_or_fail(*table_args, **table_kwargs):
        def decorator(func):
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return decorator

    def expect_all(*table_args, **table_kwargs):
        def decorator(func):
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return decorator

    def expect_all_or_drop(*table_args, **table_kwargs):
        def decorator(func):
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return decorator

    def expect_all_or_fail(*table_args, **table_kwargs):
        def decorator(func):
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return decorator

    def create_streaming_table(*args, **kwargs):
        pass


# --------------------------------------------------------------------------- #
# Wrappers                                                                    #
# --------------------------------------------------------------------------- #


def table_or_view(*args, as_view=False, **kwargs):
    if as_view:
        return view(*args, **kwargs)
    else:
        return table(*args, **kwargs)



from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.groupmember import GroupMember
from laktory.models.resources.databricks.userrole import UserRole
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class UserLookup(ResourceLookup):
    """
    Attributes
    ----------
    user_id:
         ID of the user
    """

    user_id: str = Field(serialization_alias="id")


class User(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks user

    Attributes
    ----------
    disable_as_user_deletion:
        If `True` user is disabled instead of delete when the resource is deleted
    display_name:
        Display name for the user
    group_ids:
        List of the group ids that the user should be member of.
    lookup_existing:
        Specifications for looking up existing resource. Other attributes will
        be ignored.
    roles:
        List of roles assigned to the user e.g. ("account_admin")
    workspace_access
        When `True`, the user is allowed to have workspace access

    Examples
    --------
    ```py
    from laktory import models

    u = models.resources.databricks.User(
        user_name="john.doe@okube.ai",
        display_name="John Doe",
        group_ids=[
            "${resources.group-role-engineer.id}",
            "${resources.group-domain-finance.id}",
        ],
        roles=["account_admin"],
    )
    ```
    """

    disable_as_user_deletion: bool = False
    display_name: str = None
    group_ids: list[str] = []
    lookup_existing: UserLookup = Field(None, exclude=True)
    roles: list[str] = []
    user_name: str
    workspace_access: bool = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return self.user_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - user roles
        - user group members
        """
        resources = []
        for role in self.roles:
            resources += [
                UserRole(
                    user_id=f"${{resources.{self.resource_name}.id}}",
                    role=role,
                )
            ]

        # Group Member
        for group_id in self.group_ids:
            resources += [
                GroupMember(
                    group_id=group_id,
                    member_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:User"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["groups", "roles", "group_ids"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_user"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Literal

from laktory.models.basemodel import BaseModel

PRIVILEGES = [
    "ALL_PRIVILEGES",
    "SELECT",
]


class ViewGrant(BaseModel):
    """
    Privileges granted to a principal and operating on a view

    Attributes
    ----------
    principal
        User, group or service principal name
    privileges
        List of allowed privileges

    References
    ----------
    * [privilege types](https://docs.databricks.com/en/sql/language-manual/sql-ref-privileges.html#privilege-types)
    """

    principal: str
    privileges: list[
        Literal[
            "ALL_PRIVILEGES",
            "SELECT",
        ]
    ]



import os
import shutil
from typing import Literal
from typing import Union

from pydantic import model_validator

from laktory._logger import get_logger
from laktory.models.datasinks.basedatasink import BaseDataSink
from laktory.models.datasources.tabledatasource import TableDataSource
from laktory.models.transformers.basechainnode import BaseChainNodeSQLExpr
from laktory.spark import SparkDataFrame

logger = get_logger(__name__)


class TableDataSink(BaseDataSink):
    """
    Table data sink on a metastore such as Hive, Unity Catalog or on a
    data warehouse such as Snowflake, BigQuery, etc.

    Attributes
    ----------
    checkpoint_location:
        Path to which the checkpoint file for streaming dataframe should
        be written.
    catalog_name:
        Name of the catalog of the sink table
    table_name:
        Name of the sink table
    table_type:
        Type of table. "TABLE" and "VIEW" are currently supported.
    schema_name:
        Name of the schema of the source table
    view_definition:
        View definition of "VIEW" `table_type` is selected.
    warehouse:
        Type of warehouse to which the table should be published

    Examples
    ---------
    ```python
    import pandas as pd

    from laktory import models

    df = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["AAPL", "GOOGL"],
                "price": [200.0, 205.0],
                "tstamp": ["2023-09-01", "2023-09-01"],
            }
        )
    )

    sink = models.TableDataSink(
        catalog_name="/Volumes/sources/landing/events/yahoo-finance/stock_price",
        schema_name="finance",
        table_name="slv_stock_prices",
        mode="OVERWRITE",
    )
    # sink.write(df)

    # Sink with Change Data Capture processing
    sink = models.TableDataSink(
        catalog_name="/Volumes/sources/landing/events/yahoo-finance/stock_price",
        schema_name="finance",
        table_name="slv_stock_prices",
        mode="MERGE",
        merge_cdc_options={
            "scd_type": 1,
            "primary_keys": ["symbol", "tstamp"],
        },
    )
    # sink.write(df)
    ```
    """

    catalog_name: Union[str, None] = None
    checkpoint_location: Union[str, None] = None
    format: Literal["DELTA", "PARQUET"] = "DELTA"
    schema_name: Union[str, None] = None
    table_name: Union[str, None]
    table_type: Literal["TABLE", "VIEW"] = "TABLE"
    view_definition: str = None
    warehouse: Union[Literal["DATABRICKS"], None] = "DATABRICKS"
    _parsed_view_definition: BaseChainNodeSQLExpr = None

    @model_validator(mode="after")
    def set_table_type(self):
        with self.validate_assignment_disabled():
            if self.view_definition is not None:
                self.table_type = "VIEW"
        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def full_name(self) -> str:
        """Table full name {catalog_name}.{schema_name}.{table_name}"""
        if self.table_name is None:
            return None

        name = ""
        if self.catalog_name is not None:
            name = self.catalog_name

        if self.schema_name is not None:
            if name == "":
                name = self.schema_name
            else:
                name += f".{self.schema_name}"

        if name == "":
            name = self.table_name
        else:
            name += f".{self.table_name}"

        return name

    @property
    def _id(self) -> str:
        return self.full_name

    # ----------------------------------------------------------------------- #
    # Children                                                                #
    # ----------------------------------------------------------------------- #

    @property
    def child_attribute_names(self):
        return [
            "_parsed_view_definition",
        ]

    # ----------------------------------------------------------------------- #
    # View Definition                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def parsed_view_definition(self):
        if self.view_definition is None:
            return None
        if not self._parsed_view_definition:
            self._parsed_view_definition = BaseChainNodeSQLExpr(
                expr=self.view_definition
            )
        return self._parsed_view_definition

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def _write_spark(self, df: SparkDataFrame, mode=None) -> None:
        if df.isStreaming and self._checkpoint_location is None:
            raise ValueError("Checkpoint must be provided for streaming table sink.")

        if mode is None:
            mode = self.mode

        if self.warehouse == "DATABRICKS":
            return self._write_spark_databricks(df, mode=mode)
        else:
            raise NotImplementedError(
                f"Warehouse '{self.warehouse}' is not yet supported."
            )

    def _write_spark_view(self, view_definition: str, spark) -> None:
        logger.info(f"Creating view {self.full_name} AS {view_definition}")
        df = spark.sql(f"CREATE OR REPLACE VIEW {self.full_name} AS {view_definition}")
        if self.parent_pipeline_node:
            self.parent_pipeline_node._output_df = df

    def _write_spark_databricks(self, df: SparkDataFrame, mode) -> None:
        if self.format in ["EXCEL"]:
            raise ValueError(f"'{self.format}' format is not supported with Spark")

        if mode.lower() == "merge":
            self.merge_cdc_options.execute(source=df)
            return

        # Default Options
        _options = {"mergeSchema": "true", "overwriteSchema": "false"}
        if mode in ["OVERWRITE", "COMPLETE"]:
            _options["mergeSchema"] = "false"
            _options["overwriteSchema"] = "true"
        if df.isStreaming:
            _options["checkpointLocation"] = self._checkpoint_location

        # User Options
        for k, v in self.write_options.items():
            _options[k] = v

        if df.isStreaming:
            logger.info(
                f"Writing {self._id} {self.format}  as stream with mode {mode} and options {_options}"
            )
            query = (
                df.writeStream.outputMode(mode)
                .format(self.format.lower())
                .trigger(availableNow=True)  # TODO: Add option for trigger?
                .options(**_options)
            ).toTable(self.full_name)

            query.awaitTermination()

        else:
            logger.info(
                f"Writing {self._id} {self.format}  as static with mode {mode} and options {_options}"
            )
            (
                df.write.format(self.format.lower())
                .mode(mode)
                .options(**_options)
                .saveAsTable(self.full_name)
            )

    # ----------------------------------------------------------------------- #
    # Purge                                                                   #
    # ----------------------------------------------------------------------- #

    def purge(self, spark=None):
        """
        Delete sink data and checkpoints
        """
        # Remove Data
        if self.warehouse == "DATABRICKS":
            logger.info(
                f"Dropping {self.table_type} {self.full_name}",
            )
            spark.sql(f"DROP {self.table_type} IF EXISTS {self.full_name}")

            path = self.write_options.get("path", None)
            if path and os.path.exists(path):
                is_dir = os.path.isdir(path)
                if is_dir:
                    logger.info(f"Deleting data dir {path}")
                    shutil.rmtree(path)
                else:
                    logger.info(f"Deleting data file {path}")
                    os.remove(path)
        else:
            raise NotImplementedError(
                f"Warehouse '{self.warehouse}' is not yet supported."
            )

        # Remove Checkpoint
        self._purge_checkpoint(spark=spark)

    # ----------------------------------------------------------------------- #
    # Source                                                                  #
    # ----------------------------------------------------------------------- #

    def as_source(self, as_stream=None) -> TableDataSource:
        """
        Generate a table data source with the same properties as the sink.

        Parameters
        ----------
        as_stream:
            If `True`, sink will be read as stream.

        Returns
        -------
        :
            Table Data Source
        """
        source = TableDataSource(
            catalog_name=self.catalog_name,
            table_name=self.table_name,
            schema_name=self.schema_name,
            warehouse=self.warehouse,
        )

        if as_stream:
            source.as_stream = as_stream

        source.parent = self.parent

        return source



import os
import platform
import sys

from laktory._version import VERSION


def show_version_info() -> str:
    from importlib.metadata import version

    package_names = {
        "azure-identity",
        "azure-storage-blob",
        "boto3",
        "databricks-sdk",
        "deltalake",
        "inflect",
        "mkdocs",
        "mkdocs-material",
        "mkdocs-video",
        "mkdocstrings",
        "networkx",
        "numpy",
        "pandas",
        "planck",
        "plotly",
        "polars",
        "prompt_toolkit",
        "pulumi",
        "pulumi_databricks",
        "pyarrow",
        "pydantic",
        "pyspark",
        "pytest",
        "pytest-cov",
        "pytest-examples",
        "python-dateutil",
        "pyyaml",
        "sqlparse",
        "typer",
        "typing_extensions",
        "uv",
        "yfinance",
    }
    packages = {}

    # for d in distributions():
    #     name = d.metadata["Name"]
    #     if name in package_names:
    #         packages[name] = d.version
    # packages = dict(sorted(packages.items()))

    for name in package_names:
        try:
            packages[name] = version(name)
        except:  # noqa: E722
            packages[name] = "NOT FOUND"
    packages = dict(sorted(packages.items()))

    _packages = {
        "laktory": VERSION,
        "path": os.path.abspath(os.path.dirname(__file__)),
        "python": sys.version.replace("\n", " - "),
        "platform": platform.platform(),
        "": "-------------------",
    }

    packages = {**_packages, **packages}

    info = "Laktory version info:\n"
    info += "--------------------\n"
    info += "\n".join([f"{k}-{v}" for k, v in packages.items()])

    return info



import polars as pl
from planck import units

__all__ = [
    "convert_units",
]


# --------------------------------------------------------------------------- #
# Polynomials                                                                 #
# --------------------------------------------------------------------------- #


def convert_units(
    x: pl.Expr,
    input_unit: str,
    output_unit: str,
) -> pl.Expr:
    """
    Units conversion

    Parameters
    ----------
    x:
        Input column
    input_unit:
        Input units
    output_unit:
        Output units

    Returns
    -------
    :
        Output column

    Examples
    --------
    ```py
    import polars as pl

    import laktory  # noqa: F401

    df = pl.DataFrame({"x": [1.0]})
    df = df.with_columns(
        y=pl.Expr.laktory.convert_units(pl.col("x"), input_unit="m", output_unit="ft")
    )
    print(df.glimpse(return_as_string=True))
    '''
    Rows: 1
    Columns: 2
    $ x <f64> 1.0
    $ y <f64> 3.280839895013124
    '''
    ```

    References
    ----------
    The units conversion function use [planck](https://www.okube.ai/planck/) convert as a backend.
    """
    return units.convert(x, input_unit, output_unit)



from laktory._logger import get_logger
from laktory.cli.app import app
from laktory.version import show_version_info

logger = get_logger(__name__)


@app.command()
def version():
    """
    Return installed laktory version and installed dependencies.

    Examples
    --------
    ```cmd
    laktory version
    ```
    """
    print(show_version_info())



from typing import Any
from typing import Callable
from typing import Literal
from typing import Union

from laktory._logger import get_logger
from laktory.models.transformers.basechainnode import BaseChainNode
from laktory.models.transformers.basechainnode import BaseChainNodeFuncArg
from laktory.models.transformers.basechainnode import BaseChainNodeSQLExpr
from laktory.models.transformers.basechainnode import ChainNodeColumn
from laktory.polars import PolarsDataFrame
from laktory.polars import PolarsExpr

logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Helper Classes                                                              #
# --------------------------------------------------------------------------- #


class PolarsChainNodeFuncArg(BaseChainNodeFuncArg):
    """
    Base function argument

    Attributes
    ----------
    value:
        Value of the argument
    """

    value: Union[Any]

    def eval(self):
        from laktory.models.datasources.basedatasource import BaseDataSource

        v = self.value

        if isinstance(v, BaseDataSource):
            v = self.value.read()

        elif isinstance(v, str):
            # Imports required to evaluate expressions
            import polars as pl  # noqa: F401
            from polars import col  # noqa: F401
            from polars import lit  # noqa: F401
            from polars import sql_expr  # noqa: F401

            targets = ["lit(", "col(", "sql_expr(", "pl."]

            for f in targets:
                if f in v:
                    v = eval(v)
                    break

        return v


class PolarsChainNodeSQLExpr(BaseChainNodeSQLExpr):
    """
    Chain node SQL expression

    Attributes
    ----------
    expr:
        SQL expression
    """

    def eval(self, df, chain_node=None):
        import polars as pl

        kwargs = {"df": df}
        for source in self.data_sources:
            kwargs[f"nodes__{source.node.name}"] = source.read()
        return pl.SQLContext(frames=kwargs).execute(";".join(self.parsed_expr()))


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class PolarsChainNode(BaseChainNode):
    """
    PolarsChain node that output a dataframe upon execution. As a convenience,
    `with_column` argument can be specified to create a new column from a
    polars or sql expression.

    Attributes
    ----------
    func_args:
        List of arguments to be passed to the polars function. If the function
        expects a polars expression, its string representation can be provided
        with support for `col`, `lit`, `sql_expr` and `pl.`.
    func_kwargs:
        List of keyword arguments to be passed to the polars function.If the
        function expects a polars expression, its string representation can be
        provided with support for `col`, `lit`, `sql_expr` and `pl.`.
    func_name:
        Name of the polars function to build the dataframe. Mutually
        exclusive to `sql_expr` and `with_column`.
    sql_expr:
        SQL Expression using `{df}` to reference upstream dataframe and
        defining how to build the output dataframe. Mutually exclusive to
        `func_name` and `with_column`. Other pipeline nodes can also be
        referenced using {nodes.node_name}.
    with_column:
        Syntactic sugar for adding a column. Mutually exclusive to `func_name`
        and `sql_expr`.
    with_columns:
        Syntactic sugar for adding columns. Mutually exclusive to `func_name`
        and `sql_expr`.


    Examples
    --------
    ```py
    import polars as pl

    from laktory import models

    df0 = pl.DataFrame({"x": [1, 2, 2, 3]})
    print(df0.glimpse(return_as_string=True))
    '''
    Rows: 4
    Columns: 1
    $ x <i64> 1, 2, 2, 3
    '''

    node = models.PolarsChainNode(
        with_column={"name": "cosx", "type": "double", "expr": "pl.col('x').cos()"},
    )
    df = node.execute(df0)

    node = models.PolarsChainNode(
        with_column={
            "name": "xy",
            "type": "double",
            "expr": "pl.coalesce('x')",
        },
    )
    df = node.execute(df)
    print(df.glimpse(return_as_string=True))
    '''
    Rows: 4
    Columns: 3
    $ x    <i64> 1, 2, 2, 3
    $ cosx <f64> 0.5403023058681398, -0.4161468365471424, -0.4161468365471424, -0.9899924966004454
    $ xy   <f64> 1.0, 2.0, 2.0, 3.0
    '''

    node = models.PolarsChainNode(
        func_name="unique",
        func_args=[["x"]],
        func_kwargs={"maintain_order": True},
    )
    df = node.execute(df)
    print(df.glimpse(return_as_string=True))
    '''
    Rows: 3
    Columns: 3
    $ x    <i64> 1, 2, 3
    $ cosx <f64> 0.5403023058681398, -0.4161468365471424, -0.9899924966004454
    $ xy   <f64> 1.0, 2.0, 3.0
    '''
    ```
    """

    dataframe_backend: Literal["POLARS"] = "POLARS"
    func_args: list[Union[Any]] = []
    func_kwargs: dict[str, Union[Any]] = {}
    func_name: Union[str, None] = None
    sql_expr: Union[str, None] = None
    with_column: Union[ChainNodeColumn, None] = None
    with_columns: Union[list[ChainNodeColumn], None] = []
    _parsed_func_args: list[PolarsChainNodeFuncArg] = None
    _parsed_func_kwargs: dict[str, PolarsChainNodeFuncArg] = None
    _parsed_sql_expr: PolarsChainNodeSQLExpr = None

    @property
    def parsed_func_args(self):
        if not self._parsed_func_args:
            self._parsed_func_args = [
                PolarsChainNodeFuncArg(value=a) for a in self.func_args
            ]
        return self._parsed_func_args

    @property
    def parsed_func_kwargs(self):
        if not self._parsed_func_kwargs:
            self._parsed_func_kwargs = {
                k: PolarsChainNodeFuncArg(value=v) for k, v in self.func_kwargs.items()
            }
        return self._parsed_func_kwargs

    @property
    def parsed_sql_expr(self):
        if self.sql_expr is None:
            return None
        if not self._parsed_sql_expr:
            self._parsed_sql_expr = PolarsChainNodeSQLExpr(expr=self.sql_expr)
        return self._parsed_sql_expr

    # ----------------------------------------------------------------------- #
    # Class Methods                                                           #
    # ----------------------------------------------------------------------- #

    def execute(
        self,
        df: PolarsDataFrame,
        udfs: list[Callable[[...], Union[PolarsExpr, PolarsDataFrame]]] = None,
    ) -> Union[PolarsDataFrame]:
        """
        Execute polars chain node

        Parameters
        ----------
        df:
            Input dataframe
        udfs:
            User-defined functions

        Returns
        -------
            Output dataframe
        """
        from polars import DataFrame

        from laktory.polars.datatypes import DATATYPES_MAP

        if udfs is None:
            udfs = []
        udfs = {f.__name__: f for f in udfs}

        # Build Columns
        if self._with_columns:
            for column in self._with_columns:
                logger.info(
                    f"Building column {column.name} as {column.expr or column.sql_expr}"
                )
                _col = column.eval(udfs=udfs, dataframe_backend="POLARS")
                if column.type:
                    _col = _col.cast(DATATYPES_MAP[column.type])
                df = df.with_columns(**{column.name: _col})
            return df

        # From SQL expression
        if self.sql_expr:
            logger.info(f"DataFrame {self.id} as \n{self.sql_expr.strip()}")
            return self.parsed_sql_expr.eval(df)

        # Get Function
        func_name = self.func_name
        if self.func_name is None:
            raise ValueError(
                "`func_name` must be specified if `sql_expr` is not specified"
            )

        # Get from UDFs
        f = udfs.get(func_name, None)

        # Get from built-in polars and polars extension (including Laktory) functions
        input_df = True
        if f is None:
            # Get function from namespace extension
            if "." in func_name:
                input_df = False
                vals = func_name.split(".")
                f = getattr(getattr(df, vals[0]), vals[1], None)
            else:
                f = getattr(DataFrame, func_name, None)

        if f is None:
            raise ValueError(f"Function {func_name} is not available")

        _args = self.parsed_func_args
        _kwargs = self.parsed_func_kwargs

        # Build log
        func_log = f"{func_name}("
        func_log += ",".join([a.signature() for a in _args])
        func_log += ",".join([f"{k}={a.signature()}" for k, a in _kwargs.items()])
        func_log += ")"
        logger.info(f"DataFrame {self.id} as {func_log}")

        # Build args
        args = []
        for i, _arg in enumerate(_args):
            args += [_arg.eval()]

        # Build kwargs
        kwargs = {}
        for k, _arg in _kwargs.items():
            kwargs[k] = _arg.eval()

        # Call function
        if input_df:
            df = f(df, *args, **kwargs)
        else:
            df = f(*args, **kwargs)

        return df



# COMMAND ----------
# Install dependencies

laktory_root = "/Workspace" + spark.conf.get("workspace_laktory_root", "/.laktory/")
pl_name = spark.conf.get("pipeline_name", "dlt-stock-prices")
filepath = f"{laktory_root}/pipelines/{pl_name}/requirements.txt"
# MAGIC %pip install -r $filepath
# MAGIC %restart_python

# COMMAND ----------
import importlib
import os
import sys

import pyspark.sql.functions as F  # noqa: F401

from laktory import dlt
from laktory import get_logger
from laktory import models

dlt.spark = spark
logger = get_logger(__name__)

# --------------------------------------------------------------------------- #
# Create Widgets                                                              #
# --------------------------------------------------------------------------- #

dbutils.widgets.text("pipeline_name", "dlt-stock-prices")
dbutils.widgets.text("node_name", "")
dbutils.widgets.text("workspace_laktory_root", "/.laktory/")

# --------------------------------------------------------------------------- #
# Read Pipeline                                                               #
# --------------------------------------------------------------------------- #

laktory_root = "/Workspace" + spark.conf.get(
    "workspace_laktory_root", dbutils.widgets.get("workspace_laktory_root")
)
pl_name = spark.conf.get("pipeline_name", dbutils.widgets.get("pipeline_name"))
node_name = dbutils.widgets.get("node_name")
filepath = f"{laktory_root}/pipelines/{pl_name}/config.json"
with open(filepath, "r") as fp:
    pl = models.Pipeline.model_validate_json(fp.read())

# Import User Defined Functions
sys.path.append(f"{laktory_root}/pipelines/")
udfs = []
for udf in pl.udfs:
    if udf.module_path:
        sys.path.append(os.path.abspath(udf.module_path))
    module = importlib.import_module(udf.module_name)
    module = importlib.reload(module)
    globals()[udf.module_name] = module
    udfs += [getattr(module, udf.function_name)]

# --------------------------------------------------------------------------- #
# Tables and Views Definition                                                 #
# --------------------------------------------------------------------------- #


def define_table(node, sink):
    # Get Expectations
    dlt_warning_expectations = {}
    dlt_drop_expectations = {}
    dlt_fail_expectations = {}
    if sink and not sink.is_quarantine:
        dlt_warning_expectations = node.dlt_warning_expectations
        dlt_drop_expectations = node.dlt_drop_expectations
        dlt_fail_expectations = node.dlt_fail_expectations

    # Get Name
    name = node.name
    if sink is not None:
        name = sink.table_name

    @dlt.table_or_view(
        name=name,
        comment=node.description,
        as_view=sink is None,
    )
    @dlt.expect_all(dlt_warning_expectations)
    @dlt.expect_all_or_drop(dlt_drop_expectations)
    @dlt.expect_all_or_fail(dlt_fail_expectations)
    def get_df():
        sink_str = ""
        if sink is not None:
            sink_str = f" | sink: {sink.full_name}"
        logger.info(f"Building {node.name} node{sink_str}")

        # Execute node
        node.execute(spark=spark, udfs=udfs)
        if sink and sink.is_quarantine:
            df = node.quarantine_df
        else:
            df = node.output_df
        df.printSchema()

        # Return
        return df

    return get_df


# --------------------------------------------------------------------------- #
# CDC tables                                                                  #
# --------------------------------------------------------------------------- #


def define_cdc_table(node, sink):
    dlt.create_streaming_table(
        name=sink.table_name,
        comment=node.description,
    )

    df = dlt.apply_changes(
        source=node.source.table_name, **sink.dlt_apply_changes_kwargs
    )

    return df


# --------------------------------------------------------------------------- #
# Execution                                                                   #
# --------------------------------------------------------------------------- #

# Build nodes
for node in pl.nodes:
    if node_name and node.name != node_name:
        continue

    if node.dlt_template != "DEFAULT":
        continue

    if node.sinks is None or node.sinks == []:
        wrapper = define_table(node, None)
        df = dlt.get_df(wrapper)
        display(df)

    else:
        for sink in node.sinks:
            if sink.is_cdc:
                df = define_cdc_table(node, sink)
                display(df)

            else:
                wrapper = define_table(node, sink)
                df = dlt.get_df(wrapper)
                display(df)



import io
import re
from contextlib import redirect_stdout

from pydantic import BaseModel
from pyspark.sql.connect.dataframe import DataFrame as ConnectDataFrame
from pyspark.sql.dataframe import DataFrame


class Watermark(BaseModel):
    """
    Definition of a spark structured streaming watermark for joining data
    streams.

    Attributes
    ----------
    column:
        Event time column name
    threshold:
        How late, expressed in seconds, the data is expected to be with
        respect to event time.

    References
    ----------
    https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html#handling-late-data-and-watermarking
    """

    column: str
    threshold: str


def watermark(df: DataFrame) -> Watermark:
    """
    Returns dataframe watermark if available

    Parameters
    ----------
    df:
        Input DataFrame

    Returns
    -------
    :
        Watermark column and threshold

    Examples
    --------
    ```py
    import pandas as pd

    import laktory  # noqa: F401

    df = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["AAPL", "GOOGL"],
                "price": [200.0, 205.0],
                "tstamp": ["2023-09-01", "2023-09-01"],
            }
        )
    )
    df = df.withWatermark("tstamp", "1 hour")

    watermark = df.laktory.watermark()
    ```
    """

    # Get plan
    if isinstance(df, ConnectDataFrame):
        plan = df._plan.print()

        def parse_watermark(input_string):
            match_c = re.search(r"event_time='([^']+)'", input_string)
            match_t = re.search(r"delay_threshold='([^']+)'", input_string)
            c = None
            t = None
            if match_c:
                c = match_c.group(1)
            if match_t:
                t = match_t.group(1)

            return c, t

        lines = plan.split("\n")
        for line in lines:
            if "<WithWatermark".lower() in line.lower():
                c, t = parse_watermark(line)
                return Watermark(column=c.strip(), threshold=t.strip())

    else:
        # Some Databricks cluster types prevent from using private method
        # like ._jdf. Instead, we use .explain() method and capture the
        # output
        with io.StringIO() as buf, redirect_stdout(buf):
            df.explain(extended=True)
            plan = buf.getvalue().strip()

        lines = plan.split("\n")
        for line in lines:
            if "EventTimeWatermark".lower() in line.lower():
                c, t = line.lower().replace("'eventtimewatermark '", "").split(",")
                return Watermark(column=c.strip(), threshold=t.strip())



import polars as pl

from laktory._logger import get_logger

logger = get_logger(__name__)


def smart_join(
    left: pl.DataFrame,
    other: pl.DataFrame,
    how: str = "left",
    on: list[str] = None,
    left_on: list[str] = None,
    other_on: list[str] = None,
    coalesce: bool = False,
) -> pl.DataFrame:
    """
    Join tables and coalesce join columns. Optionally coalesce columns found
    in both left and other not used in join.

    Parameters
    ----------
    left:
        Left side of the join
    other:
        Right side of the join
    how:
        Type of join (left, outer, full, etc.)
    on:
        A list of strings for the columns to join on. The columns must exist
        on both sides.
    left_on:
        Name(s) of the left join column(s).
    other_on:
        Name(s) of the right join column(s).
    coalesce:
        If `True` columns present in both left and other are coalesced.
        Columns part of the join are always coalesced.

    Examples
    --------
    ```py
    import polars as pl

    import laktory  # noqa: F401

    df_prices = pl.DataFrame(
        {
            "symbol": ["AAPL", "GOOGL"],
            "price": [200.0, 205.0],
        }
    )

    df_meta = pl.DataFrame(
        {
            "symbol": ["AAPL", "GOOGL"],
            "name": ["Apple", "Google"],
        }
    )

    df = df_prices.laktory.smart_join(
        other=df_meta,
        on=["symbol"],
    )

    print(df.glimpse(return_as_string=True))
    '''
    Rows: 2
    Columns: 3
    $ symbol <str> 'AAPL', 'GOOGL'
    $ price  <f64> 200.0, 205.0
    $ name   <str> 'Apple', 'Google'
    '''
    ```
    """
    logger.info(
        f"Executing {left.laktory.signature()} {how} JOIN {other.laktory.signature()}"
    )

    # Validate inputs
    if left_on or other_on:
        if not other_on:
            raise ValueError("If `left_on` is set, `other_on` should also be set")
        if not left_on:
            raise ValueError("If `other_on` is set, `left_on` should also be set")
    if not (on or left_on or other_on):
        raise ValueError("Either `on` or (`left_on` and `other_on`) should be set")

    # Parse inputs
    if isinstance(left_on, str):
        left_on = [left_on]
    if isinstance(other_on, str):
        other_on = [other_on]

    # Drop duplicates to prevent adding rows to left
    if on:
        other = other.unique(on, maintain_order=True)

    if on:
        logger.info(f"   ON {on}")
    else:
        logger.info(f"   ON left.{left_on} == other.{other_on}")

    logger.debug(f"Left Schema: {left.laktory.signature()}")
    logger.debug(f"Other Schema: {other.laktory.signature()}")

    suffix = "_right"
    df = left.join(
        other=other,
        on=on,
        left_on=left_on,
        right_on=other_on,
        suffix=suffix,
        how=how,
        coalesce=True,  # to be consistent with Spark smart_join
    )

    if coalesce:
        for c1 in df.columns:
            if c1.endswith(suffix):
                c0 = c1.replace(suffix, "")
                df = df.with_columns(**{c0: pl.coalesce(pl.col(c0), pl.col(c1))})
                df = df.drop(c1)

    # Drop watermark column
    logger.debug(f"Joined Schema: {df.schema}")

    return df


if __name__ == "__main__":
    import pandas as pd

    from laktory._testing.stockprices import spark

    df_prices = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["AAPL", "GOOGL"],
                "price": [200.0, 205.0],
            }
        )
    )

    df_meta = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["AAPL", "GOOGL"],
                "name": ["Apple", "Google"],
            }
        )
    )

    df = df_prices.smart_join(
        other=df_meta,
        on=["symbol"],
    )



from collections import defaultdict

from pyspark.sql.dataframe import DataFrame

from laktory._logger import get_logger
from laktory.spark.dataframe.watermark import Watermark
from laktory.spark.dataframe.watermark import watermark

logger = get_logger(__name__)


def smart_join(
    left: DataFrame,
    other: DataFrame,
    how: str = "left",
    on: list[str] = None,
    on_expression: str = None,
    left_on: list[str] = None,
    left_watermark: Watermark = None,
    other_on: list[str] = None,
    other_watermark: Watermark = None,
    time_constraint_interval_lower: str = "60 seconds",
    time_constraint_interval_upper: str = None,
    coalesce: bool = False,
) -> DataFrame:
    """
     Join tables and coalesce join columns. Optionally coalesce columns found
    in both left and other not used in join. Support for watermarking for
    streaming joins.

    Parameters
    ----------
    left:
        Left side of the join
    other:
        Right side of the join
    how:
        Type of join (left, outer, full, etc.)
    on:
        A list of strings for the columns to join on. The columns must exist
        on both sides.
    on_expression:
        String expression the join on condition. The expression can include
        `left` and `other` dataframe references.
    left_on:
        Name(s) of the left join column(s).
    left_watermark
        Watermark for left dataframe
    other_on:
        Name(s) of the right join column(s).
    other_watermark
        Watermark for other dataframe
    time_constraint_interval_lower:
        Lower bound for a spark streaming event-time constraint
    time_constraint_interval_upper:
        Upper bound for a spark streaming event-time constraint
    coalesce:
        If `True` columns present in both left and other are coalesced.

    Examples
    --------
    ```py
    import pandas as pd

    import laktory  # noqa: F401

    df_prices = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["AAPL", "GOOGL"],
                "price": [200.0, 205.0],
            }
        )
    )

    df_meta = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["AAPL", "GOOGL"],
                "name": ["Apple", "Google"],
            }
        )
    )

    df = df_prices.laktory.smart_join(
        other=df_meta,
        on=["symbol"],
    )

    print(df.toPandas().to_string())
    '''
       price    name symbol
    0  200.0   Apple   AAPL
    1  205.0  Google  GOOGL
    '''
    ```

    References
    ----------

    * [pyspark join](https://spark.apache.org/docs/3.1.2/api/python/reference/api/pyspark.sql.DataFrame.join.html)
    * [spark streaming join](https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html#inner-joins-with-optional-watermarking)
    """
    import pyspark.sql.functions as F

    logger.info(f"Executing {left} {how} JOIN {other}")

    # Validate inputs
    if left_on or other_on:
        if not other_on:
            raise ValueError("If `left_on` is set, `other_on` should also be set")
        if not left_on:
            raise ValueError("If `other_on` is set, `left_on` should also be set")
    if not (on or left_on or on_expression):
        raise ValueError(
            "Either `on` or (`left_on` and `other_on`) or `on_expression` should be set"
        )

    # Parse inputs
    if on is None:
        on = []
    elif isinstance(on, str):
        on = [on]
    if left_on is None:
        left_on = []
    elif isinstance(left_on, str):
        left_on = [left_on]
    if other_on is None:
        other_on = []
    elif isinstance(other_on, str):
        other_on = [other_on]

    wml = left_watermark
    wmo = other_watermark
    if wml is not None and not isinstance(wml, Watermark):
        wml = Watermark(**wml)
    if wmo is not None and not isinstance(wmo, Watermark):
        wmo = Watermark(**wmo)

    # Set watermarks
    if wml is None:
        try:
            wml = watermark(left)
        except Exception as e:
            logger.warn(f"Could not fetch wartermark from left dataframe: {e}")
    else:
        left = left.withWatermark(wml.column, wml.threshold)
    if wmo is None:
        try:
            wmo = watermark(other)
        except Exception as e:
            logger.warn(f"Could not fetch wartermark from other dataframe: {e}")
    else:
        other = other.withWatermark(wmo.column, wmo.threshold)

    # Add watermark
    other_cols = []
    if wmo is not None:
        other_cols += [F.col(wmo.column).alias("_other_wc")]
    other_cols += [F.col(c) for c in other.columns]
    other = other.select(other_cols)

    # Drop duplicates to prevent adding rows to left
    if on:
        other = other.dropDuplicates(on)

    _join = []
    for c in on:
        _join += [f"left.{c} == other.{c}"]
    for l, o in zip(left_on, other_on):
        _join += [f"left.{l} == other.{o}"]
    if on_expression:
        _join += [on_expression]

    if wmo is not None:
        if time_constraint_interval_lower:
            _join += [
                f"left.{wml.column} >= other._other_wc - interval {time_constraint_interval_lower}"
            ]
        if time_constraint_interval_upper:
            _join += [
                f"left.{wml.column} <= other._other_wc + interval {time_constraint_interval_upper}"
            ]
    _join = " AND ".join(_join)

    logger.info(f"   ON {_join}")
    logger.debug(f"Left Schema: {left.schema}")
    logger.debug(f"Other Schema: {other.schema}")

    df = left.alias("left").join(
        other=other.alias("other"),
        on=F.expr(_join),
        how=how,
    )

    # Find duplicated columns (because of join)
    d = defaultdict(lambda: 0)
    for c in df.columns:
        d[c] += 1

    # Drop duplicated columns
    for c, v in d.items():
        if v < 2:
            continue

        if not (coalesce or c in _join):
            continue
        df = df.withColumn("__tmp", F.coalesce(f"left.{c}", f"other.{c}"))
        df = df.drop(c)
        df = df.withColumn(c, F.col("__tmp"))
        df = df.drop("__tmp")

    # Drop watermark column
    if wmo is not None:
        df = df.drop(F.col("other._other_wc"))
    logger.debug(f"Joined Schema: {df.schema}")

    return df


if __name__ == "__main__":
    import pandas as pd

    from laktory._testing.stockprices import spark

    df_prices = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["AAPL", "GOOGL"],
                "price": [200.0, 205.0],
            }
        )
    )

    df_meta = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["AAPL", "GOOGL"],
                "name": ["Apple", "Google"],
            }
        )
    )

    df = df_prices.smart_join(
        other=df_meta,
        on=["symbol"],
    )



from .accesscontrol import AccessControl
from .alert import Alert
from .catalog import Catalog
from .cluster import Cluster
from .clusterpolicy import ClusterPolicy
from .dashboard import Dashboard
from .dbfsfile import DbfsFile
from .directory import Directory
from .dltpipeline import DLTPipeline
from .externallocation import ExternalLocation
from .grants import Grants
from .group import Group
from .job import Job
from .metastore import Metastore
from .metastoreassignment import MetastoreAssignment
from .metastoredataaccess import MetastoreDataAccess
from .mlflowexperiment import MLflowExperiment
from .mlflowmodel import MLflowModel
from .mlflowwebhook import MLflowWebhook
from .mwsnccbinding import MwsNccBinding
from .mwsnetworkconnectivityconfig import MwsNetworkConnectivityConfig
from .mwspermissionassignment import MwsPermissionAssignment
from .notebook import Notebook
from .permissions import Permissions
from .query import Query
from .repo import Repo
from .schema import Schema
from .secret import Secret
from .secretacl import SecretAcl
from .secretscope import SecretScope
from .serviceprincipal import ServicePrincipal
from .serviceprincipalrole import ServicePrincipalRole
from .table import Table
from .user import User
from .userrole import UserRole
from .vectorsearchendpoint import VectorSearchEndpoint
from .vectorsearchindex import VectorSearchIndex
from .volume import Volume
from .warehouse import Warehouse
from .workspacefile import WorkspaceFile



from typing import Literal
from typing import Union

from laktory._logger import get_logger
from laktory.models.datasources.basedatasource import BaseDataSource
from laktory.spark import SparkDataFrame

logger = get_logger(__name__)


class TableDataSource(BaseDataSource):
    """
    Data source using a data warehouse data table, generally used in the
    context of a data pipeline. Currently only supported by Spark dataframes.

    Attributes
    ----------
    catalog_name:
        Name of the catalog of the source table
    schema_name:
        Name of the schema of the source table
    table_name:
        Name of the source table

    Examples
    ---------
    ```python
    from laktory import models

    source = models.TableDataSource(
        catalog_name="dev",
        schema_name="finance",
        table_name="brz_stock_prices",
        selects=["symbol", "open", "close"],
        filter="symbol='AAPL'",
        as_stream=True,
    )
    # df = source.read(spark)
    ```
    """

    catalog_name: Union[str, None] = None
    table_name: Union[str, None] = None
    schema_name: Union[str, None] = None
    warehouse: Union[Literal["DATABRICKS"], None] = "DATABRICKS"

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def full_name(self) -> str:
        """Table full name {catalog_name}.{schema_name}.{table_name}"""
        if self.table_name is None:
            return None

        name = ""
        if self.catalog_name is not None:
            name = self.catalog_name

        if self.schema_name is not None:
            if name == "":
                name = self.schema_name
            else:
                name += f".{self.schema_name}"

        if name == "":
            name = self.table_name
        else:
            name += f".{self.table_name}"

        return name

    @property
    def _id(self) -> str:
        return self.full_name

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _read_spark(self, spark) -> SparkDataFrame:
        if self.warehouse == "DATABRICKS":
            return self._read_spark_databricks(spark)
        else:
            raise NotImplementedError(
                f"Warehouse '{self.warehouse}' is not yet supported."
            )

    def _read_spark_databricks(self, spark) -> SparkDataFrame:
        if self.as_stream:
            logger.info(f"Reading {self._id} as stream")
            df = spark.readStream.table(self.full_name)
        else:
            logger.info(f"Reading {self._id} as static")
            df = spark.read.table(self.full_name)

        return df
